theory Abstract_Rewriting_Impl
imports "$AFP/Abstract-Rewriting/Abstract_Rewriting"
begin

partial_function (option) compute_NF :: "('a \<Rightarrow> 'a option) \<Rightarrow> 'a \<Rightarrow> 'a option"
  where [simp,code]: "compute_NF f a = (case f a of None \<Rightarrow> Some a | Some b \<Rightarrow> compute_NF f b)"

lemma compute_NF_sound: assumes res: "compute_NF f a = Some b"
  and f_sound: "\<And> a b. f a = Some b \<Longrightarrow> (a,b) \<in> r"
  shows "(a,b) \<in> r^*"
proof (induct rule: compute_NF.raw_induct[OF _ res, of "\<lambda> g a b. g = f \<longrightarrow> (a,b) \<in> r^*", THEN mp[OF _ refl]])
  case (1 cnf g a b)
  show ?case
  proof
    assume "g = f"
    note 1 = 1[unfolded this]
    show "(a,b) \<in> r^*"
    proof (cases "f a")
      case None
      with 1(2) show ?thesis by simp
    next
      case (Some c)
      from 1(2)[unfolded this] have "cnf f c = Some b" by simp
      from 1(1)[OF this] have "(c,b) \<in> r^*" by auto
      with f_sound[OF Some] show ?thesis by auto
    qed
  qed
qed

lemma compute_NF_complete: assumes res: "compute_NF f a = Some b"
  and f_complete: "\<And> a. f a = None \<Longrightarrow> a \<in> NF r"
  shows "b \<in> NF r"
proof (induct rule: compute_NF.raw_induct[OF _ res, of "\<lambda> g a b. g = f \<longrightarrow> b \<in> NF r", THEN mp[OF _ refl]])
  case (1 cnf g a b)
  show ?case
  proof
    assume "g = f"
    note 1 = 1[unfolded this]
    show "b \<in> NF r"
    proof (cases "f a")
      case None
      with f_complete[OF None] 1(2)
      show ?thesis by simp
    next
      case (Some c)
      from 1(2)[unfolded this] have "cnf f c = Some b" by simp
      from 1(1)[OF this] show ?thesis by simp
    qed
  qed
qed

lemma compute_NF_SN: assumes SN: "SN r"
  and f_sound: "\<And> a b. f a = Some b \<Longrightarrow> (a,b) \<in> r"
  shows "\<exists> b. compute_NF f a = Some b" (is "?P a")
proof -
  let ?r = "{(a,b). f a = Some b}"
  have "?r \<subseteq> r" using f_sound by auto
  from SN_subset[OF SN this] have SNr: "SN ?r" .
  show ?thesis
  proof (induct rule: SN_induct[OF SNr, of "\<lambda> a. ?P a"])
    case (1 a)
    show ?case
    proof (cases "f a")
      case None thus ?thesis by auto
    next
      case (Some b)
      hence "(a,b) \<in> ?r" by simp
      from 1[OF this] f_sound[OF Some] show ?thesis 
        using Some by auto
    qed
  qed
qed

end