(*
Copyright 2012 René Thiemann

This file is part of IsaFoR/CeTA.

IsaFoR/CeTA is free software: you can redistribute it and/or modify it under the
terms of the GNU Lesser General Public License as published by the Free Software
Foundation, either version 3 of the License, or (at your option) any later
version.

IsaFoR/CeTA is distributed in the hope that it will be useful, but WITHOUT ANY
WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License along
with IsaFoR/CeTA. If not, see <http://www.gnu.org/licenses/>.
*)
theory Bounded_Increase
imports 
  QDP_Framework 
  Name 
  Signature_Extension 
  Countables 
  Non_Inf_Order 
  Substitution 
  Dependency_Graph
begin

fun invert_dep :: "dependance \<Rightarrow> dependance" where 
  "invert_dep Increase = Decrease"
| "invert_dep Decrease = Increase"  
| "invert_dep d = d"
  

lemma rel_dep_mono: "r1 \<subseteq> r2 \<Longrightarrow> r1 ^^^ k \<subseteq> r2 ^^^ k"
  by (cases k, auto)

lemma rel_dep_invert_mono: "r1^-1 \<subseteq> r2 \<Longrightarrow> r1 ^^^ invert_dep k \<subseteq> r2 ^^^ k"
  by (cases k, auto)


fun compat_root where 
  "compat_root _ (Var _) = False"
| "compat_root (Var _) l = False"
| "compat_root fls gts = (root fls = root gts)"


locale fixed_trs_dep =
  fixes \<pi> :: "'f dep"
  and R :: "('f,'v)trs"
  and Q :: "('f,'v)term set"
  and nfs :: bool
begin

abbreviation NFQ where "NFQ \<equiv> NF_terms Q"

inductive gen_usable_rule :: "('f,'v)term \<Rightarrow> ('f,'v)rule \<Rightarrow> bool"
where in_R: "(l,r) \<in> R \<Longrightarrow> compat_root l t \<Longrightarrow> gen_usable_rule t (l,r)"
   |  in_U: "(l,r) \<in> R \<Longrightarrow> compat_root l t \<Longrightarrow> gen_usable_rule r lr \<Longrightarrow> gen_usable_rule t lr"
   |  in_arg: "i < length ts \<Longrightarrow> gen_usable_rule (ts ! i) lr \<Longrightarrow> lr' \<in> {lr} ^^^ \<pi> (f,length ts) i \<Longrightarrow> gen_usable_rule (Fun f ts) lr'"

definition gen_usable_rules :: "('f,'v)term \<Rightarrow> ('f,'v)trs"
  where "gen_usable_rules t \<equiv> {lr. gen_usable_rule t lr}"
  
definition gen_usable_rules_pairs :: "('f,'v)trs \<Rightarrow> ('f,'v)trs"
  where "gen_usable_rules_pairs P \<equiv> \<Union> gen_usable_rules ` snd ` P"

lemma gen_usable_rulesI[intro]: "gen_usable_rule t lr \<Longrightarrow> lr \<in> gen_usable_rules t"
  unfolding gen_usable_rules_def by auto

lemma gen_usable_rulesE[elim]: "lr \<in> gen_usable_rules t \<Longrightarrow> gen_usable_rule t lr"
  unfolding gen_usable_rules_def by auto

lemma gen_usable_rules_Fun: "gen_usable_rules (Fun f ts) = 
  {(Fun f ls,r) | ls r. (Fun f ls,r) \<in> R \<and> length ls = length ts} \<union>
  \<Union> { gen_usable_rules r | ls r. (Fun f ls,r) \<in> R \<and> length ls = length ts} \<union>
  \<Union> { gen_usable_rules (ts ! i) ^^^ \<pi> (f,length ts) i | i. i < length ts}" (is "?L = ?R1 \<union> ?R2 \<union> ?R3")
proof (rule equalityI) 
  let ?R = "?R1 \<union> ?R2 \<union> ?R3"
  {
    fix l r
    assume "(l,r) \<in> ?L"
    hence "gen_usable_rule (Fun f ts) (l,r)" by auto
    hence "(l,r) \<in> ?R" 
    proof(cases)
      case in_R
      from in_R(2) obtain ls where l: "l = Fun f ls" by (cases l, auto)
      with in_R show ?thesis by auto
    next
      case (in_U l2 r2)
      from in_U(2) obtain ls where l2: "l2 = Fun f ls" by (cases l2, auto)
      with in_U show ?thesis by auto
    next
      case (in_arg i lr)
      hence "(l, r) \<in> {lr. gen_usable_rule (ts ! i) lr} ^^^ \<pi> (f, length ts) i"
        by (cases "\<pi> (f,length ts) i", auto)
      thus ?thesis using in_arg(1) unfolding gen_usable_rules_def by auto
    qed
  } 
  thus "?L \<subseteq> ?R" ..
  moreover
  {
    fix l r
    assume "(l,r) \<in> ?R"
    hence "(l,r) \<in> ?R1 \<or> (l,r) \<in> ?R2 \<or> (l,r) \<in> ?R3" by blast
    hence "(l,r) \<in> ?L"
    proof
      assume "(l,r) \<in> ?R1"
      with in_R[of l r] show "(l,r) \<in> ?L" by auto
    next
      assume "(l,r) \<in> ?R2 \<or> (l,r) \<in> ?R3"
      thus ?thesis
      proof
        assume "(l,r) \<in> ?R2"
        then obtain ls r2 where U: "(l,r) \<in> gen_usable_rules r2" 
          and R: "(Fun f ls, r2) \<in> R" and l: "length ls = length ts" by auto
        from in_U[OF R _ gen_usable_rulesE[OF U]] l
        show ?thesis by auto
      next
        assume "(l,r) \<in> ?R3"
        then obtain i where i: "i < length ts"
          and lr: "(l,r) \<in> gen_usable_rules (ts ! i) ^^^ \<pi> (f,length ts) i" by auto
        then obtain lr' where u: "gen_usable_rule (ts ! i) lr'" and lr: "(l,r) \<in> {lr'} ^^^ \<pi> (f,length ts) i"
          unfolding gen_usable_rules_def by (cases "\<pi> (f,length ts) i", auto)
        show "(l,r) \<in> ?L"
          by (intro gen_usable_rulesI, rule in_arg[OF i u lr])
      qed
    qed
  } 
  thus "?R \<subseteq> ?L" ..
qed

lemma gen_usable_rules_Var: "gen_usable_rules (Var x) = {}"
proof -
  {
    fix l r
    assume "(l,r) \<in> gen_usable_rules (Var x)"
    hence "gen_usable_rule (Var x) (l,r)" ..
    hence False by (cases, auto)
  }
  thus ?thesis by auto
qed  
    
lemma gen_usable_rules_subst: "gen_usable_rules t \<subseteq> gen_usable_rules (t \<cdot> \<sigma>)"
proof (induct t)
  case (Var x)
  thus ?case by (simp add: gen_usable_rules_Var)
next
  case (Fun f ts)
  {
    fix i
    assume i: "i < length ts"
    hence "ts ! i \<in> set ts" by auto
    from Fun[OF this] i
    have "gen_usable_rules (ts ! i) ^^^ \<pi> (f, length ts) i \<subseteq> 
      gen_usable_rules (map (\<lambda>t. t \<cdot> \<sigma>) ts ! i) ^^^ \<pi> (f, length ts) i" (is "?l \<subseteq> _")
      by (intro rel_dep_mono, auto)
    also have "... \<subseteq> \<Union>{gen_usable_rules (map (\<lambda>t. t \<cdot> \<sigma>) ts ! i) ^^^ \<pi> (f, length ts) i |i. i < length ts}"
      (is "_ \<subseteq> ?r")
      using i by auto
    finally have "?l \<subseteq> ?r" .
  } note main = this
  show ?case unfolding subst_apply_term.simps gen_usable_rules_Fun
    by (insert main, auto) 
qed

lemma gen_usable_rules_split: "\<exists> k. \<forall> s t. gen_usable_rules s ^^^ k \<subseteq> gen_usable_rules t ^^^ k \<longrightarrow>
  gen_usable_rules (Fun f (b @ s # a)) \<subseteq> gen_usable_rules (Fun f (b @ t # a))" 
proof -
  let ?i = "length b"
  let ?n = "Suc (?i + length a)"
  let ?\<pi> = "\<pi> (f,?n) ?i"
  let ?all = "\<lambda> t. gen_usable_rules (Fun f (b @ t # a))"
  let ?coll = "\<lambda> t. \<Union> (\<lambda> i. gen_usable_rules ((b @ t # a) ! i) ^^^ \<pi> (f,?n) i) ` {i . i < ?n}"
  let ?single = "\<lambda> t. gen_usable_rules t ^^^ \<pi> (f,?n) ?i"
  have id: "{i. i < ?n} = {i. i < ?i} \<union> {?i} \<union> {Suc ?i + k | k. Suc ?i + k < ?n}" (is "_ = ?ib \<union> ?ii \<union> ?ia") by (auto, presburger)
  { 
    fix s t
    have "?all s - ?all t \<subseteq> ?coll s - ?coll t" 
      unfolding gen_usable_rules_Fun by auto
    also have "\<dots> \<subseteq> ?single s - ?single t" unfolding id by (auto simp: nth_append)
    finally have "?all s - ?all t \<subseteq> ?single s - ?single t" .
  } note sub = this
  have "\<exists> k. \<forall> s t. gen_usable_rules s ^^^ k \<subseteq> gen_usable_rules t ^^^ k \<longrightarrow> ?single s - ?single t = {}"
  proof (cases ?\<pi>)
    case Ignore
    thus ?thesis by auto
  next
    case Increase
    show ?thesis
      by (rule exI[of _ Increase], insert Increase, auto)
  next
    case Decrease
    show ?thesis
      by (rule exI[of _ Decrease], insert Decrease, auto)
  next
    case Wild
    show ?thesis
      by (rule exI[of _ Wild], insert Wild, auto)
  qed
  then obtain k where "\<And> s t. gen_usable_rules s ^^^ k \<subseteq> gen_usable_rules t ^^^ k \<Longrightarrow> ?single s - ?single t = {}" by auto
  with sub have "\<And> s t. gen_usable_rules s ^^^ k \<subseteq> gen_usable_rules t ^^^ k \<Longrightarrow> ?all s \<subseteq> ?all t" by blast
  thus "\<exists> k. \<forall> s t. gen_usable_rules s ^^^ k \<subseteq> gen_usable_rules t ^^^ k \<longrightarrow> ?all s \<subseteq> ?all t" by auto
qed
end

locale fixed_trs_dep_order = fixed_trs_dep \<pi> R Q nfs + non_inf_order S NS F \<pi>
  for \<pi> :: "('f :: countable) dep"
  and R :: "('f,'v :: countable)trs"
  and Q :: "('f,'v)term set"
  and F :: "'f sig"
  and S :: "('f,'v)trs"
  and NS :: "('f,'v)trs"
  and nfs :: bool
  and m :: bool
  + fixes const :: 'f
  assumes wwf: "wwf_qtrs Q R"
  and NFQ: "NF_terms Q \<subseteq> NF_trs R"
  and infinite_V: "infinite (UNIV :: 'v set)"
  and RF: "funas_trs R \<subseteq> F"
  and QF: "\<Union> funas_term ` Q \<subseteq> F"
begin

lemma main_technical_lemma: 
  assumes step: "(t \<cdot> \<sigma>,v) \<in> qrstep nfs Q R"
  and \<sigma>: "\<sigma> ` vars_term t \<subseteq> NFQ"
  and tsigF: "funas_term (t \<cdot> \<sigma>) \<subseteq> F \<or> funas_args_term (t \<cdot> \<sigma>) \<subseteq> F \<and> \<not> defined (applicable_rules Q R) (the (root (t \<cdot> \<sigma>)))"
  shows "(gen_usable_rules t ^^^ k \<subseteq> NS \<longrightarrow> {(t \<cdot> \<sigma>, v)} ^^^ k \<subseteq> NS) \<and> (\<exists> u \<sigma>'. v = u \<cdot> \<sigma>' 
    \<and> gen_usable_rules u ^^^ k \<subseteq> gen_usable_rules t ^^^ k \<and> \<sigma>' ` vars_term u \<subseteq> NFQ)"
  using step \<sigma> tsigF
proof (induction t arbitrary: v k)
  case (Var x v k)
  from Var.prems(2) NFQ have "Var x \<cdot> \<sigma> \<in> NF_trs R" by auto
  moreover from Var.prems(1) have "(Var x \<cdot> \<sigma>, v) \<in> rstep R" by auto
  ultimately have False by auto
  thus ?case by auto
next
  case (Fun f ts v k)
  note switch = wwf_qtrs_imp_nfs_switch[OF wwf]
  from qrstepE[OF Fun.prems(1)[unfolded switch[of nfs True]]]
    obtain C \<sigma>' l r where
    nf: "\<forall>u\<lhd>l \<cdot> \<sigma>'. u \<in> NFQ" and 
    lr: "(l, r) \<in> R" and 
    t: "Fun f ts \<cdot> \<sigma> = C\<langle>l \<cdot> \<sigma>'\<rangle>" and 
    v: "v = C\<langle>r \<cdot> \<sigma>'\<rangle>" and 
    nfs: "NF_subst True (l, r) \<sigma>' Q" .
  let ?\<sigma>' = "map (\<lambda> t. t \<cdot> \<sigma>')"
  let ?\<sigma> = "map (\<lambda> t. t \<cdot> \<sigma>)"
  from wwf_qtrs_imp_left_fun[OF wwf lr] obtain g ls where l: "l = Fun g ls" by auto
  from wwf only_applicable_rules[OF nf] lr
  have vars_lr: "vars_term r \<subseteq> vars_term l" unfolding wwf_qtrs_def by auto
  from RF lr have rF: "funas_term r \<subseteq> F" unfolding funas_trs_def funas_rule_def by force
  show ?case
  proof (cases C)
    case Hole
    from t Hole have t: "Fun f ts \<cdot> \<sigma> = l \<cdot> \<sigma>'" by simp
    from only_applicable_rules[OF nf, of r] lr l
    have defg: "defined (applicable_rules Q R) (g,length ls)" unfolding defined_def applicable_rules_def 
      by force
    from v Hole have v: "v = r \<cdot> \<sigma>'" by simp
    from l t have gf: "g = f" and id: "?\<sigma> ts = ?\<sigma>' ls" by auto
    from arg_cong[OF id, of length] have len: "length ts = length ls" by simp
    with l t gf have compat: "compat_root l (Fun f ts)" by simp
    from gf defg len have "defined (applicable_rules Q R) (f,length ts)" by simp
    with Fun(4) have "funas_term (Fun f ts \<cdot> \<sigma>) \<subseteq> F" by auto
    with t have "funas_term (l \<cdot> \<sigma>') \<subseteq> F" by simp
    hence \<sigma>'F: "\<Union> funas_term ` \<sigma>' ` vars_term l \<subseteq> F" unfolding funas_term_subst by blast
    let ?\<sigma>' = "\<lambda> x. if x \<in> vars_term l then \<sigma>' x else Var x"
    have l_id: "l \<cdot> \<sigma>' = l \<cdot> ?\<sigma>'" by (rule term_subst_eq, simp)
    have r_id: "r \<cdot> \<sigma>' = r \<cdot> ?\<sigma>'" by (rule term_subst_eq, insert vars_lr, auto)
    have \<sigma>'F: "\<Union> funas_term ` range ?\<sigma>' \<subseteq> F" using \<sigma>'F by auto
    {
      assume ns: "gen_usable_rules (Fun f ts) ^^^ k \<subseteq> NS"
      have "(l,r) \<in> gen_usable_rules (Fun f ts)"
        by (rule gen_usable_rulesI[OF in_R[OF lr compat]])
      hence "{(l,r)} \<subseteq> gen_usable_rules (Fun f ts)" by auto
      from subset_trans[OF rel_dep_mono[OF this] ns]
      have "{(l,r)} ^^^ k \<subseteq> NS" .
      hence "{(l \<cdot> \<sigma>', r \<cdot> \<sigma>')} ^^^ k \<subseteq> (\<lambda> (s,t). (s \<cdot> ?\<sigma>', t \<cdot> ?\<sigma>')) ` NS" 
        unfolding l_id r_id by (cases k, auto)
      also have "(\<lambda> (s,t). (s \<cdot> ?\<sigma>', t \<cdot> ?\<sigma>')) ` NS \<subseteq> NS" using F_subst_closedE[OF stable_NS \<sigma>'F] by auto
      finally have "{(Fun f ts \<cdot> \<sigma>, v)} ^^^ k \<subseteq> NS" using t v by simp
    }
    hence NS: "gen_usable_rules (Fun f ts) ^^^ k \<subseteq> NS \<longrightarrow> {(Fun f ts \<cdot> \<sigma>, v)} ^^^ k \<subseteq> NS" ..
    show ?thesis
    proof (rule conjI[OF NS], intro exI conjI)
      show "v = r \<cdot> ?\<sigma>'" using v r_id by simp
      show "gen_usable_rules r ^^^ k \<subseteq> gen_usable_rules (Fun f ts) ^^^ k"
        by (rule rel_dep_mono, insert in_U[OF lr compat], auto simp: gen_usable_rules_def)
      show "?\<sigma>' ` vars_term r \<subseteq> NFQ" using vars_lr nfs unfolding NF_subst_def vars_rule_def by auto
    qed
  next
    case (More g bef D aft)
    let ?i' = "length bef"
    let ?tsi' = "ts ! ?i'"
    let ?dl = "D \<langle> l \<cdot> \<sigma>' \<rangle>"
    let ?dr = "D \<langle> r \<cdot> \<sigma>' \<rangle>"
    let ?n' = "Suc (?i' + length aft)"
    from t More have t: "Fun f ts \<cdot> \<sigma> = Fun f (bef @ ?dl # aft)" and gf: "g = f" by auto
    with v More have v: "v = Fun f (bef @ ?dr # aft)" by auto
    from t have "map (\<lambda> t. t \<cdot> \<sigma>) ts = bef @ ?dl # aft" by simp
    from arg_cong[OF this, of length] have len: "length ts = ?n'" by simp
    hence i': "?i' < length ts" by auto
    from id_take_nth_drop[OF i'] obtain tbef taft where ts: "Fun f ts = Fun f (tbef @ ?tsi' # taft)" and tbef: "tbef = take ?i' ts" by auto
    let ?i = "length tbef" let ?n = "Suc (?i + length taft)"
    let ?tsi = "ts ! ?i"
    from tbef i' have llen: "?i' = ?i" by auto
    note ts = ts[unfolded llen]
    note tbef = tbef[unfolded llen]
    note i = i'[unfolded llen]
    from i have mem: "?tsi \<in> set ts" by auto
    from Fun.prems(2) mem have vars: "\<sigma> ` vars_term ?tsi \<subseteq> NFQ" by auto
    from t[unfolded ts] have "?\<sigma> tbef @ (?tsi \<cdot> \<sigma>) # ?\<sigma> taft = bef @ ?dl # aft" by auto
    from this[unfolded append_eq_append_conv_if] 
    have bef: "bef = ?\<sigma> tbef" and tsi: "?tsi \<cdot> \<sigma> =  ?dl"
     and aft: "aft = ?\<sigma> taft" using llen by auto
    from Fun(4) have tF: "\<And> t. t \<in> set ts \<Longrightarrow> funas_term (t \<cdot> \<sigma>) \<subseteq> F" by (auto simp: funas_args_term_def)
    from tF[OF mem] have dlF: "funas_term ?dl \<subseteq> F" unfolding tsi .
    with vars_lr rF have drF: "funas_term ?dr \<subseteq> F" by (force simp: funas_term_subst)
    have "(?tsi \<cdot> \<sigma>,?dr) \<in> qrstep True Q R" unfolding tsi
      by (rule qrstepI[OF nf lr _ _ nfs], auto)
    note IH = Fun.IH[OF mem this[unfolded switch[of True nfs]] vars disjI1[OF tF[OF mem]]]
    from Fun(4)[unfolded t bef aft] dlF have fdlF: "funas_args_term (Fun f (?\<sigma> tbef @ ?dl # ?\<sigma> taft)) \<subseteq> F" by (auto simp: funas_args_term_def)
    from Fun(4)[unfolded t bef aft] drF have fdrF: "funas_args_term (Fun f (?\<sigma> tbef @ ?dr # ?\<sigma> taft)) \<subseteq> F" by (auto simp: funas_args_term_def)
    note compat' = dep_compatE[OF dep_compat_NS, of f "?\<sigma> tbef" _ "?\<sigma> taft"]
    note compat_lr = compat'[OF fdlF drF]
    note compat_rl = compat'[OF fdrF dlF]
    note v = v[unfolded bef aft]
    have t: "Fun f ts \<cdot> \<sigma> = Fun f (?\<sigma> tbef @ ?dl # ?\<sigma> taft)" unfolding ts tsi[symmetric]
      by simp
    {
      assume pi: "\<pi> (f,?n) ?i = Increase \<or> \<pi> (f,?n) ?i = Wild"
      have "gen_usable_rules ?tsi ^^^ k \<subseteq> gen_usable_rules (Fun f ts) ^^^ k"
      proof (rule rel_dep_mono, rule, intro gen_usable_rulesI)
        fix l r
        assume "(l, r) \<in> gen_usable_rules ?tsi"
        note lr = gen_usable_rulesE[OF this]
        show "gen_usable_rule (Fun f ts) (l, r)" unfolding ts
          by (rule in_arg[of ?i _ "(l,r)"], insert lr pi, auto)
      qed
    } note inc = this
    {
      assume pi: "\<pi> (f,?n) ?i = Decrease \<or> \<pi> (f,?n) ?i = Wild"
      have "gen_usable_rules ?tsi ^^^ invert_dep k \<subseteq> gen_usable_rules (Fun f ts) ^^^ k"
      proof (rule rel_dep_invert_mono, rule, intro gen_usable_rulesI)
        fix l r
        assume "(l, r) \<in> (gen_usable_rules ?tsi)^-1"
        hence "(r,l) \<in> gen_usable_rules ?tsi" by auto
        note lr = gen_usable_rulesE[OF this]
        show "gen_usable_rule (Fun f ts) (l, r)" unfolding ts
          by (rule in_arg[of ?i _ "(r,l)"], insert lr pi, auto)
      qed
    } note dec = this
    {
      assume prems: "gen_usable_rules (Fun f ts) ^^^ k \<subseteq> NS" 
      note us = this[unfolded ts]
      have orient: "{(Fun f ts \<cdot> \<sigma>, v)} ^^^ k \<subseteq> NS"
      proof (cases "\<pi> (f,?n) ?i")
        case Ignore
        with compat_lr compat_rl
        show ?thesis unfolding t v by (cases k, auto)
      next
        case Increase
        have "gen_usable_rules ?tsi ^^^ k \<subseteq> gen_usable_rules (Fun f ts) ^^^ k"
          by (rule inc, insert Increase, auto)
        from subset_trans[OF this prems] have "gen_usable_rules ?tsi ^^^ k \<subseteq> NS" .
        from IH this have "{(?dl,?dr)} ^^^ k \<subseteq> NS" unfolding tsi by simp
        with compat_lr compat_rl Increase show ?thesis unfolding t v 
          by (cases k, auto)
      next
        case Decrease
        have "gen_usable_rules ?tsi ^^^ invert_dep k \<subseteq> gen_usable_rules (Fun f ts) ^^^ k"
          by (rule dec, insert Decrease, auto)
        from subset_trans[OF this prems] have "gen_usable_rules ?tsi ^^^ invert_dep k \<subseteq> NS" .
        from IH this have "{(?dl,?dr)} ^^^ invert_dep k \<subseteq> NS" unfolding tsi by simp
        with compat_lr compat_rl Decrease show ?thesis unfolding t v 
          by (cases k, auto)
      next
        case Wild
        have one: "gen_usable_rules ?tsi ^^^ invert_dep k \<subseteq> gen_usable_rules (Fun f ts) ^^^ k"
          by (rule dec, insert Wild, auto)
        have two: "gen_usable_rules ?tsi ^^^ k \<subseteq> gen_usable_rules (Fun f ts) ^^^ k"
          by (rule inc, insert Wild, auto)
        from one two have "k = Ignore \<or> gen_usable_rules ?tsi ^^^ Wild \<subseteq> gen_usable_rules (Fun f ts) ^^^ k"
          by (cases k, auto)
        thus ?thesis
        proof
          assume "k = Ignore"
          thus ?thesis by auto
        next
          assume "gen_usable_rules ?tsi ^^^ Wild \<subseteq> gen_usable_rules (Fun f ts) ^^^ k"
          from subset_trans[OF this prems] have "gen_usable_rules ?tsi ^^^ Wild \<subseteq> NS" .
          from IH[THEN conjunct1, rule_format, OF this] have "{(?dl,?dr)} ^^^ Wild \<subseteq> NS" unfolding tsi by auto
          with compat_lr compat_rl Wild show ?thesis unfolding t v 
            by (cases k, auto)
        qed
      qed
    }
    hence orient: "gen_usable_rules (Fun f ts) ^^^ k \<subseteq> NS \<longrightarrow> {(Fun f ts \<cdot> \<sigma>, v)} ^^^ k \<subseteq> NS" ..
    let ?ts = "\<lambda> t. tbef @ t # taft"
    let ?C = "\<lambda> t. Fun f (?ts t)"
    {
      fix j
      from IH[THEN conjunct2]
      obtain u \<mu> where dr: "?dr = u \<cdot> \<mu>"
        and us: "gen_usable_rules u ^^^ j \<subseteq> gen_usable_rules ?tsi ^^^ j"
        and nf: "\<mu> ` vars_term u \<subseteq> NFQ"
        by blast
      have "finite (vars_term (?C u))" by auto
      from finite_fresh_names_infinite_univ[OF this infinite_V]
        obtain ren ren' where
        ren: "\<And> x. x \<in> vars_term (?C u) \<Longrightarrow> ren x \<notin> vars_term (?C u)"
        and ren': "\<And> x. x \<in> vars_term (?C u) \<Longrightarrow> ren' (ren x) = x" by blast
      let ?ren = "(\<lambda> x. Var (ren x)) :: ('f,'v)subst"
      let ?ren' = "(\<lambda> x. Var (ren' x)) :: ('f,'v)subst"
      let ?u = "u \<cdot> ?ren"
      let ?\<mu> = "\<lambda> x. if (x \<in> vars_term (?C u)) then \<sigma> x else (?ren' \<circ>s \<mu>) x"
      have "\<exists>u \<mu>. v = ?C u \<cdot> \<mu> \<and> gen_usable_rules u ^^^ j \<subseteq> gen_usable_rules ?tsi ^^^ j \<and> \<mu> ` vars_term (?C u) \<subseteq> NFQ"
      proof (intro exI conjI)
        have "?dr = u \<cdot> (?ren \<circ>s ?\<mu>)" unfolding dr
          by (rule term_subst_eq, 
          unfold subst_compose_def,
          insert ren ren', auto)
        also have "... = ?u \<cdot> ?\<mu>" by simp
        finally have "?dr = ?u \<cdot> ?\<mu>" .
        moreover
        have "?\<sigma> (tbef @ taft) = map (\<lambda> t. t \<cdot> ?\<mu>) (tbef @ taft)"
        proof (rule map_cong[OF refl])
          fix t 
          assume t: "t \<in> set (tbef @ taft)"
          show "t \<cdot> \<sigma> = t \<cdot> ?\<mu>"
            by (rule term_subst_eq, insert t, auto)
        qed
        ultimately
        show "v = ?C ?u \<cdot> ?\<mu>" unfolding v by simp
      next
        have "gen_usable_rules ?u \<subseteq> gen_usable_rules (?u \<cdot> ?ren')"
          by (rule gen_usable_rules_subst)
        also have "?u \<cdot> ?ren' = u \<cdot> (?ren \<circ>s ?ren')" by simp
        also have "... = u \<cdot> Var"
          by (rule term_subst_eq, unfold subst_compose_def, insert ren ren', auto)
        finally have us': "gen_usable_rules ?u \<subseteq> gen_usable_rules u" by simp
        show "gen_usable_rules ?u ^^^ j \<subseteq> gen_usable_rules ?tsi ^^^ j"
          by (rule subset_trans[OF rel_dep_mono[OF us'] us])
      next
        {
          fix x
          assume "x \<in> vars_term (?C ?u)"
          hence "x \<in> vars_term (Fun f (tbef @ taft)) \<or> x \<in> vars_term ?u" by auto
          hence "?\<mu> x \<in> NFQ"
          proof
            assume "x \<in> vars_term (Fun f (tbef @ taft))"
            with Fun.prems(2) show ?thesis unfolding ts by auto
          next
            assume "x \<in> vars_term ?u"
            then obtain y where y: "y \<in> vars_term u" and x: "x = ren y" 
              unfolding vars_term_subst by auto
            with nf ren ren' show ?thesis unfolding subst_compose_def by auto
          qed
        }
        thus "?\<mu> ` vars_term (?C ?u) \<subseteq> NFQ" by blast
      qed          
    } 
    note IH = this
    show ?thesis
    proof (rule conjI[OF orient])
      from gen_usable_rules_split[of f tbef taft] obtain j where 
        usi: "\<And> s t. gen_usable_rules s ^^^ j \<subseteq> gen_usable_rules t ^^^ j \<Longrightarrow> gen_usable_rules (Fun f (tbef @ s # taft)) \<subseteq> gen_usable_rules (Fun f (tbef @ t # taft))"
        by auto
      from IH[of j] obtain u \<mu> where v: "v = ?C u \<cdot> \<mu>"
        and usi_prem: "gen_usable_rules u ^^^ j \<subseteq> gen_usable_rules ?tsi ^^^ j"
        and nf: "\<mu> ` vars_term (?C u) \<subseteq> NFQ" by auto
      note usi = usi[OF usi_prem]
      show "\<exists>u \<mu>. v = u \<cdot> \<mu> \<and> gen_usable_rules u ^^^ k \<subseteq> gen_usable_rules (Fun f ts) ^^^ k \<and> \<mu> ` vars_term u \<subseteq> NFQ"
        by (rule exI[of _ "?C u"], rule exI[of _ \<mu>], intro conjI[OF v conjI[OF _ nf]], unfold ts,
        rule rel_dep_mono[OF usi])
    qed
  qed
qed

lemma gen_usable_rules_lemma: 
  assumes step: "(t \<cdot> \<sigma>,v) \<in> (qrstep nfs Q R)^*"
  and usable: "gen_usable_rules t \<subseteq> NS"
  and \<sigma>: "\<sigma> ` vars_term t \<subseteq> NFQ"
  and tsigF: "funas_args_term (t \<cdot> \<sigma>) \<subseteq> F"
  and ndef: "\<not> defined (applicable_rules Q R) (the (root (t \<cdot> \<sigma>)))"
  shows "(t \<cdot> \<sigma>, v) \<in> NS"
proof -
  let ?R = "qrstep nfs Q R"
  from rtrancl_imp_relpow[OF step] obtain n
    where "(t \<cdot> \<sigma>, v) \<in> ?R^^n" ..
  from this \<sigma> usable tsigF ndef show ?thesis
  proof (induction n arbitrary: t \<sigma> v)
    case 0
    thus ?case using refl_NS unfolding refl_on_def by simp
  next
    case (Suc n)
    from Suc.prems(1)
    have "(t \<cdot> \<sigma>, v) \<in> ?R^^(1 + n)" by auto
    from this[unfolded relpow_add] obtain u where
      one: "(t \<cdot> \<sigma>, u) \<in> ?R" and n: "(u,v) \<in> ?R ^^ n" by auto
    from Suc.prems(4-5) have "funas_args_term (t \<cdot> \<sigma>) \<subseteq> F \<and> \<not> defined (applicable_rules Q R) (the (root (t \<cdot> \<sigma>)))" by auto
    from main_technical_lemma[OF one Suc.prems(2) disjI2[OF this], of Increase] Suc.prems(3)
    obtain s \<mu> where ns: "(t \<cdot> \<sigma>, u) \<in> NS" and u: "u = s \<cdot> \<mu>"
      and us: "gen_usable_rules s \<subseteq> gen_usable_rules t"
      and nf: "\<mu> ` vars_term s \<subseteq> NFQ"
      by auto
    from qrstep_imp_nrqrstep[OF _ Suc.prems(5) one[unfolded u]]
    have one: "(t \<cdot> \<sigma>, s \<cdot> \<mu>) \<in> nrqrstep nfs Q R" using wwf_qtrs_imp_left_fun[OF wwf] by force
    from nrqrstep_preserves_root[OF one] Suc.prems(5) have ndef: "\<not> defined (applicable_rules Q R) (the (root (s \<cdot> \<mu>)))" by simp
    from nrqrstep_funas_args_term[OF wwf RF Suc.prems(4) one] have Fs: "funas_args_term (s \<cdot> \<mu>) \<subseteq> F" .
    from Suc.IH[OF n[unfolded u] nf _ Fs ndef] us Suc.prems(3) have "(s \<cdot> \<mu>, v) \<in> NS" by auto
    with ns trans_NS show ?case unfolding u trans_def by blast
  qed
qed

lemma compat_rtrancl: assumes ab: "(a,b) \<in> S"
  and bc: "(b,c) \<in> (NS \<union> S)^*"
  shows "(a,c) \<in> S"
  using bc
proof (induct)
  case base
  show ?case by (rule ab)
next
  case (step c d)
  from step(2-3) show ?case using compat trans_S unfolding trans_def by blast
qed
end

datatype ('f,'v)cond_constraint 
  = CC_cond bool "('f,'v)rule" 
  | CC_rewr "('f,'v)term" "('f,'v)term"
  | CC_impl "('f,'v)cond_constraint list" "('f,'v)cond_constraint"
  | CC_all 'v "('f,'v)cond_constraint"

lemma
  fixes P :: "('f, 'v) cond_constraint \<Rightarrow> bool"
  assumes "\<And> ct s t. P(CC_cond ct (s,t))" and
  "\<And> s t. P(CC_rewr s t)" and 
  "\<And>cs c. (\<And>c. c \<in> set cs \<Longrightarrow> P c) \<Longrightarrow> P c \<Longrightarrow> P(CC_impl cs c)" and
  "\<And> x c. P c \<Longrightarrow> P(CC_all x c)"
  shows cond_constraint_induct[case_names cond rewr impl all, induct type: cond_constraint]:
    "P c" and "\<And>c. c \<in> set cs \<Longrightarrow> P c"
using assms by (induct c and cs, auto) 

fun concl_of :: "('f,'v)cond_constraint \<Rightarrow> ('f,'v)cond_constraint" where
  "concl_of (CC_impl c1 c2) = c2"
| "concl_of c = c"

fun prems_of :: "('f,'v)cond_constraint \<Rightarrow> ('f,'v)cond_constraint list" where
  "prems_of (CC_impl c1 c2) = c1"
| "prems_of c = []"

definition normalize_cc :: "('f,'v)cond_constraint \<Rightarrow> ('f,'v)cond_constraint" where
  "normalize_cc c \<equiv> CC_impl (prems_of c) (concl_of c)"


fun vars_cc :: "('f,'v)cond_constraint \<Rightarrow> 'v set" where
  "vars_cc (CC_cond ct (s,t)) = vars_term s \<union> vars_term t"
| "vars_cc (CC_rewr s t) = vars_term s \<union> vars_term t"
| "vars_cc (CC_impl cs c) = (\<Union> vars_cc ` set cs) \<union> vars_cc c"
| "vars_cc (CC_all x c) = vars_cc c - {x}"

fun vars_cc_list :: "('f,'v)cond_constraint \<Rightarrow> 'v list" where
  "vars_cc_list (CC_cond ct (s,t)) = vars_term_list s @ vars_term_list t"
| "vars_cc_list (CC_rewr s t) = vars_term_list s @ vars_term_list t"
| "vars_cc_list (CC_impl c1 c2) = concat (map vars_cc_list c1) @ vars_cc_list c2"
| "vars_cc_list (CC_all x c) = [ y . y <- vars_cc_list c, y \<noteq> x]"

lemma vars_cc_list[simp]: "set (vars_cc_list c) = vars_cc c" by (induct c, auto)

lemma vars_cc_normalize[simp]: "vars_cc (normalize_cc c) = vars_cc c"
  by (cases c, auto simp: normalize_cc_def)

lemma finite_vars_cc: "finite (vars_cc c)" unfolding vars_cc_list[symmetric] by (rule finite_set)

fun cc_bound :: "'v list \<Rightarrow> ('f,'v)cond_constraint \<Rightarrow> ('f,'v)cond_constraint" where
  "cc_bound [] c = c"
| "cc_bound (x # xs) c = CC_all x (cc_bound xs c)"

fun cc_unbound :: "('f,'v)cond_constraint \<Rightarrow> 'v list \<times> ('f,'v)cond_constraint" where
  "cc_unbound (CC_all x c) = (case cc_unbound c of (xs,d) \<Rightarrow> (x # xs,d))"
| "cc_unbound c = ([],c)"

lemma cc_unbound_bound: "cc_unbound c = (xs,d) \<Longrightarrow> cc_bound xs d = c"
proof (induction c arbitrary: xs d)
  case (all x c xs d)
  obtain ys e where c: "cc_unbound c = (ys,e)" by force
  with all(2) have xs: "xs = x # ys" and e: "e = d" by auto
  from all(1)[OF c[unfolded e]] have "cc_bound ys d = c" .
  thus ?case unfolding xs by auto
qed auto

definition disjoint_variant :: "('f,'v)rule list \<Rightarrow> ('f,'v)rule list \<Rightarrow> bool"
  where "disjoint_variant sts uvs \<equiv> 
    length sts = length uvs \<and> 
    (\<forall> i < length sts. sts ! i =v uvs ! i) \<and> 
    is_partition (map vars_rule uvs)"

definition initial_conditions_gen :: "('a \<Rightarrow> 'a \<Rightarrow> bool) \<Rightarrow> nat \<Rightarrow> nat \<Rightarrow> 'a set \<Rightarrow> 'a \<Rightarrow> 'a list set" where
  "initial_conditions_gen p bef_len aft_len P st \<equiv> let
     pairs = (\<lambda> n. { sts. length sts = n \<and> set sts \<subseteq> P});
     all = { bef @ st # aft | bef aft. bef \<in> pairs bef_len \<and> aft \<in> pairs aft_len}
   in { bef_st_aft . bef_st_aft \<in> all \<and> (\<forall> i < bef_len + aft_len. p (bef_st_aft ! i) (bef_st_aft ! Suc i))}"

lemma initial_conditions_gen_mono: assumes "\<And> s t. p s t \<Longrightarrow> q s t"
  shows "initial_conditions_gen p bef_len aft_len P st \<subseteq> initial_conditions_gen q bef_len aft_len P st"
  using assms unfolding initial_conditions_gen_def Let_def by auto

datatype condition_type = Bound | Strict | Non_Strict

fun condition_of :: "'f \<Rightarrow> condition_type \<Rightarrow> ('f,'v)rule \<Rightarrow> ('f,'v)cond_constraint" where
  "condition_of c Bound (s,_) = CC_cond False (s,Fun c [])"
| "condition_of c Strict st = CC_cond True st"
| "condition_of c Non_Strict st = CC_cond False st"

definition constraint_of :: "'f \<Rightarrow> condition_type \<Rightarrow> ('f,'v)rule list \<Rightarrow> nat \<Rightarrow> ('f,'v)cond_constraint" where
  "constraint_of c ctype uvs bef \<equiv> 
  CC_impl 
    (map (\<lambda> i. CC_rewr (snd (uvs ! i)) (fst (uvs ! Suc i))) [0 ..< length uvs - 1])
    (condition_of c ctype (uvs ! bef))"

context fixed_trs_dep_order
begin

definition rel_of :: "bool \<Rightarrow> ('f,'v)trs" where
  "rel_of ct \<equiv> if ct then S else NS"

definition normal_F_subst :: "('f,'v)subst \<Rightarrow> bool"
  where "normal_F_subst \<sigma> \<equiv> (\<Union> funas_term ` range \<sigma> \<subseteq> F) \<and> range \<sigma> \<subseteq> NF_trs R"

definition m_SN :: "('f,'v)term \<Rightarrow> bool" where
  "m_SN t \<equiv> m \<longrightarrow> SN_on (qrstep nfs Q R) {t}"

fun cc_models :: "('f,'v)subst \<Rightarrow> ('f,'v)cond_constraint \<Rightarrow> bool" (infix "\<Turnstile>" 51) where
  "\<sigma> \<Turnstile> CC_cond ct (s,t) = ((s \<cdot> \<sigma>,t \<cdot> \<sigma>) \<in> rel_of ct)"
| "\<sigma> \<Turnstile> CC_impl c1 c2 = (Ball (set c1) (cc_models \<sigma>) \<longrightarrow> \<sigma> \<Turnstile> c2)"
| "\<sigma> \<Turnstile> CC_rewr s t = (m_SN (s \<cdot> \<sigma>) \<and> (s \<cdot> \<sigma>, t \<cdot> \<sigma>) \<in> (qrstep nfs Q R)^* \<and> t \<cdot> \<sigma> \<in> NF_terms Q)"
| "\<sigma> \<Turnstile> CC_all x c = (\<forall> (t :: ('f,'v)term). funas_term t \<subseteq> F \<longrightarrow> t \<in> NF_trs R \<longrightarrow> (\<sigma> (x := t)) \<Turnstile> c)"

lemma normalize_cc: "(\<sigma> \<Turnstile> c) = (\<sigma> \<Turnstile> normalize_cc c)"  unfolding normalize_cc_def
proof (cases c)
  case (CC_impl c1 c2)
  show "\<sigma> \<Turnstile> c = \<sigma> \<Turnstile> CC_impl ((prems_of c)) (concl_of c)" unfolding CC_impl by auto
qed auto

definition cc_valid :: "('f,'v)cond_constraint \<Rightarrow> bool" where
  "cc_valid c \<equiv> \<forall> \<sigma>. normal_F_subst \<sigma> \<longrightarrow> \<sigma> \<Turnstile> c"

lemma cc_validE[elim]: "cc_valid c \<Longrightarrow> normal_F_subst \<sigma> \<Longrightarrow> \<sigma> \<Turnstile> c" unfolding cc_valid_def by blast
lemma cc_validI[intro!]: "\<lbrakk>\<And> \<sigma>. normal_F_subst \<sigma> \<Longrightarrow> \<sigma> \<Turnstile> c\<rbrakk> \<Longrightarrow> cc_valid c" unfolding cc_valid_def by blast

lemma normalize_cc_valid: "cc_valid (normalize_cc c) \<Longrightarrow> cc_valid c" 
  unfolding cc_valid_def normalize_cc[symmetric] by auto

definition cc_equiv :: "('f,'v)cond_constraint \<Rightarrow> ('f,'v)cond_constraint \<Rightarrow> bool" (infix "=cc" 49) where
  "c =cc d \<equiv> \<forall> \<sigma>. \<sigma> \<Turnstile> c = \<sigma> \<Turnstile> d"

definition cc_implies :: "('f,'v)cond_constraint \<Rightarrow> ('f,'v)cond_constraint \<Rightarrow> bool" (infix "\<longrightarrow>cc" 49) where
  "c \<longrightarrow>cc d \<equiv> \<forall> \<sigma>. \<sigma> \<Turnstile> c \<longrightarrow> \<sigma> \<Turnstile> d"

lemma cc_equiv_substE: "c =cc d \<Longrightarrow> \<sigma> \<Turnstile> c = \<sigma> \<Turnstile> d" unfolding cc_equiv_def by auto
lemma cc_equivI[intro]: "\<lbrakk>\<And> \<sigma>. \<sigma> \<Turnstile> c = \<sigma> \<Turnstile> d\<rbrakk> \<Longrightarrow> c =cc d" unfolding cc_equiv_def by auto
lemma cc_impliesI[intro]: "\<lbrakk>\<And> \<sigma>. \<sigma> \<Turnstile> c \<Longrightarrow> \<sigma> \<Turnstile> d\<rbrakk> \<Longrightarrow> c \<longrightarrow>cc d" unfolding cc_implies_def by auto
lemma cc_impliesE[elim]: "c \<longrightarrow>cc d \<Longrightarrow> cc_valid c \<Longrightarrow> cc_valid d" unfolding cc_implies_def cc_valid_def by auto
lemma cc_implies_substE[elim]: "c \<longrightarrow>cc d \<Longrightarrow> \<sigma> \<Turnstile> c \<Longrightarrow> \<sigma> \<Turnstile> d" unfolding cc_implies_def cc_valid_def by auto
lemma cc_equiv_validE: "c =cc d \<Longrightarrow> cc_valid c = cc_valid d" unfolding cc_valid_def cc_equiv_def by auto
lemma cc_equiv_refl[simp]: "c =cc c" by auto
lemma cc_implies_refl[simp]: "c \<longrightarrow>cc c" by auto
lemma normalize_cc_equivI: "normalize_cc c = normalize_cc d \<Longrightarrow> c =cc d"
  by (rule cc_equivI, unfold normalize_cc[of _ c] normalize_cc[of _ d], auto)

definition initial_conditions :: "nat \<Rightarrow> nat \<Rightarrow> ('f,'v)trs \<Rightarrow> ('f,'v)rule \<Rightarrow> ('f,'v)rule list set" where
  "initial_conditions bef aft P \<equiv> initial_conditions_gen (\<lambda> st uv. (st,uv) \<in> DG nfs m P Q R) bef aft P"

definition constraint_present :: "nat \<Rightarrow> nat \<Rightarrow> ('f,'v)trs \<Rightarrow>
   condition_type \<Rightarrow> ('f,'v)cond_constraint set \<Rightarrow> ('f,'v)rule \<Rightarrow> bool" where
  "constraint_present bef aft P ctype ccs st \<equiv>
     (\<forall> sts. sts \<in> initial_conditions bef aft P st \<longrightarrow> 
     (\<exists> c uvs. disjoint_variant sts uvs \<and> c \<in> ccs \<and> c \<longrightarrow>cc constraint_of const ctype uvs bef))"

fun tcondition_of :: "condition_type \<Rightarrow> ('f,'v)rule \<Rightarrow> bool" where
  "tcondition_of Bound (s,t) = ((s,Fun const []) \<in> NS)"
| "tcondition_of Strict st = (st \<in> S)"
| "tcondition_of Non_Strict st = (st \<in> NS)"

lemma tcondition_of: 
  assumes model: "\<sigma> \<Turnstile> condition_of const ctype (s,t)"
  shows "tcondition_of ctype (s \<cdot> \<sigma>,t \<cdot> \<sigma>)"
  using assms by (cases ctype, auto simp: rel_of_def)

lemma conditional_general_reduction_pair_chains: 
  assumes U_NS: "gen_usable_rules_pairs (P \<union> Pw) \<subseteq> NS"
  and chain: "min_ichain_sig (nfs,m,P,Pw,Q,Rs,Rw) F s t \<sigma>"
  and nvar: "\<And> s t. (s,t) \<in> P \<union> Pw \<Longrightarrow> is_Fun s \<and> is_Fun t"
  and var_cond: "\<And> s t. (s,t) \<in> P \<union> Pw \<Longrightarrow> vars_term t \<subseteq> vars_term s"
  and ndef: "\<And> s t. (s,t) \<in> P \<union> Pw \<Longrightarrow> \<not> defined (applicable_rules Q (Rs \<union> Rw)) (the (root t))"
  and PF: "\<And> s t. (s,t) \<in> P \<union> Pw \<Longrightarrow> funas_args_term t \<subseteq> F"
  and orient: "P \<union> Pw \<subseteq> Ps \<union> Pns"
  and cS: "Ball Ps (constraint_present bef aft (P \<union> Pw) Strict ccs)"
  and cNS: "Ball Pns (constraint_present bef aft (P \<union> Pw) Non_Strict ccs)"
  and cB: "Ball Pb (constraint_present bef aft (P \<union> Pw) Bound ccs)"
  and ccs: "Ball ccs cc_valid"
  and R: "Rs \<union> Rw = R"
  shows "\<exists> i. min_ichain_sig (nfs,m,P - Ps, Pw - Ps,Q,Rs,Rw) F (shift s i) (shift t i) (shift \<sigma> i) 
    \<or> min_ichain_sig (nfs,m,P - Pb, Pw - Pb,Q,Rs,Rw) F (shift s i) (shift t i) (shift \<sigma> i)"
proof -
  let ?DP = "(nfs,m,P,Pw,Q,Rs,Rw)"
  let ?DPm = "\<lambda> S. (nfs,m,P - S,Pw - S,Q,Rs,Rw)"
  let ?DP_Ps = "?DPm Ps"
  let ?DP_Pb = "?DPm Pb"
  show ?thesis
  proof (cases "(INFM i. (s i, t i) \<in> Ps) \<and> (INFM i. (s i, t i) \<in> Pb)")
    case False
    then obtain i where disj: "(\<forall> j \<ge> i. (s j, t j) \<notin> Ps) \<or> (\<forall> j \<ge> i. (s j, t j) \<notin> Pb)"
      unfolding INFM_nat_le by blast
    let ?s = "shift s"
    let ?t = "shift t"
    let ?\<sigma> = "shift \<sigma>"
    from disj
    show ?thesis
    proof
      assume nS: "(\<forall> j \<ge> i. (s j, t j) \<notin> Ps)"
      let ?DP = "(nfs, m, P - Ps, Pw - Ps, Q, Rs - {}, Rw - {})"
      have "\<exists> i. min_ichain_sig ?DP F (?s i) (?t i) (?\<sigma> i)"
      proof (rule min_ichain_split_sig[OF chain], rule)
        assume "min_ichain_sig (nfs, m, Ps \<inter> (P \<union> Pw), P \<union> Pw - Ps, Q, {} \<inter> (Rs \<union> Rw), Rs \<union> Rw - {}) F s t \<sigma>"
        hence "INFM i. (s i, t i) \<in> Ps \<inter> (P \<union> Pw)" by (auto simp: ichain.simps)
        with nS show False unfolding INFM_nat_le by blast
      qed
      thus ?thesis by auto
    next
      assume nB: "(\<forall> j \<ge> i. (s j, t j) \<notin> Pb)"
      let ?DP = "(nfs, m, P - Pb, Pw - Pb, Q, Rs - {}, Rw - {})"
      have "\<exists> i. min_ichain_sig ?DP F (?s i) (?t i) (?\<sigma> i)"
      proof (rule min_ichain_split_sig[OF chain], rule)
        assume "min_ichain_sig (nfs, m, Pb \<inter> (P \<union> Pw), P \<union> Pw - Pb, Q, {} \<inter> (Rs \<union> Rw), Rs \<union> Rw - {}) F s t \<sigma>"
        hence "INFM i. (s i, t i) \<in> Pb \<inter> (P \<union> Pw)" by (auto simp: ichain.simps)
        with nB show False unfolding INFM_nat_le by blast
      qed
      thus ?thesis by auto
    qed
  next
    case True
    let ?QR = "qrstep nfs Q R"
    let ?ndef = "\<lambda> t. \<not> defined (applicable_rules Q R) (the (root t))"
    from chain have ichain: "ichain ?DP s t \<sigma>" and \<sigma>F: "funas_ichain s t \<sigma> \<subseteq> F" by auto
    {
      fix i
      from \<sigma>F have "\<Union> funas_term ` range (\<sigma> i) \<subseteq> F" unfolding funas_ichain_def by auto
    } note \<sigma>F = this
    let ?st = "\<lambda> i. (s i, t i)"
    from ichain have P_Pw: "\<And> i. ?st i \<in> P \<union> Pw" unfolding ichain.simps by auto
    from ichain have NF: "\<And> i. s i \<cdot> \<sigma> i \<in> NFQ" unfolding ichain.simps by auto
    {
      fix ctype PP i      
      assume i_bef: "i \<ge> bef" and mem: "(s i, t i) \<in> PP"
      assume cp: "Ball PP (constraint_present bef aft (P \<union> Pw) ctype ccs)"
      with mem have cp: "constraint_present bef aft (P \<union> Pw) ctype ccs (s i, t i)" by auto
      def j \<equiv> "i - bef" 
      def n \<equiv> "Suc (bef + aft)"
      let ?j = "\<lambda> i. i + j"
      from i_bef j_def have i_j_bef: "i = ?j bef" by auto
      let ?sts_gen = "\<lambda> j. map (\<lambda> i. ?st (i + j))"
      let ?sts = "?sts_gen j [0 ..< n]"
      let ?bef = "?sts_gen j [0 ..< bef]"
      let ?aft = "?sts_gen (j + Suc bef) [0 ..< aft]"
      have "?sts = ?sts_gen j [0 ..< bef + Suc 0 + aft]" unfolding n_def by auto
      also have "... = ?bef @ ?st i # ?aft"
        unfolding map_upt_add i_j_bef by (simp add: ac_simps)
      finally have sts: "?sts = ?bef @ ?st i # ?aft" .
      have sts_init: "?sts \<in> initial_conditions bef aft (P \<union> Pw) (s i, t i)"
        unfolding initial_conditions_def initial_conditions_gen_def Let_def
      proof (rule, rule, rule, intro exI conjI, rule sts, insert P_Pw, (auto)[2], intro allI impI)
        fix i
        assume "i < bef + aft"
        with n_def have i: "i < n" "Suc i < n" by auto
        hence sts: "(?sts ! i, ?sts ! Suc i) = (?st (?j i), ?st (Suc (?j i)))" by auto
        show "(?sts ! i, ?sts ! Suc i) \<in> DG nfs m (P \<union> Pw) Q R" unfolding sts
          by (rule DG_I[of _ _ _ _ _ "\<sigma> (?j i)" "\<sigma> (Suc (?j i))"], 
            insert chain R, auto simp: ichain.simps minimal_cond_def)
      qed
      from cp[unfolded constraint_present_def, rule_format, OF sts_init]
      obtain c uvs where var: "disjoint_variant ?sts uvs" and c_ccs: "c \<in> ccs" and c_uvs: "c \<longrightarrow>cc constraint_of const ctype uvs bef" by auto
      def us \<equiv> "\<lambda> i. fst (uvs ! i)"
      def vs \<equiv> "\<lambda> i. snd (uvs ! i)"
      let ?uv = "\<lambda> i. (us i, vs i)"
      from cc_impliesE[OF c_uvs] ccs c_ccs
      have valid: "cc_valid (constraint_of const ctype uvs bef)" by auto
      note var = var[unfolded disjoint_variant_def]
      from var have part: "is_partition (map vars_rule uvs)" and ulen: "length uvs = n" by auto
      {
        fix i
        assume "i < n"
        with var have "?st (?j i) =v ?uv i" unfolding us_def vs_def by auto
        from eq_rule_mod_varsE[OF this] this have
        "\<exists> \<tau>. s (?j i) = us i \<cdot> \<tau> \<and> t (?j i) = vs i \<cdot> \<tau> \<and> range \<tau> \<subseteq> range Var" "?st (?j i) =v ?uv i" by auto
      } 
      hence eq: "\<And> i. i < n \<Longrightarrow> ?st (?j i) =v ?uv i" and taus: "\<forall> i. (\<exists> \<tau>. i < n \<longrightarrow> s (?j i) = us i \<cdot> \<tau> \<and> t (?j i) = vs i \<cdot> \<tau> \<and> range \<tau> \<subseteq> range Var)" by blast+
      from choice[OF taus] obtain \<tau> where st_uv: "\<And> i. i < n \<Longrightarrow> s (?j i) = us i \<cdot> \<tau> i" "\<And> i. i < n \<Longrightarrow> t (?j i) = vs i \<cdot> \<tau> i" and \<tau>: "\<And> i. i < n \<Longrightarrow> range (\<tau> i) \<subseteq> range Var" by auto
      let ?sigs = "map (\<lambda> i. \<tau> i \<circ>s \<sigma> (?j i)) [0 ..< n]"
      def \<gamma> \<equiv> "\<lambda> x. if x \<in> \<Union> vars_rule ` ?uv ` {0 ..< n} then fun_merge ?sigs (map vars_rule uvs) x else Var x"
      {
        fix i x
        assume i: "i < n" and x: "x \<in> vars_rule (uvs ! i)"
        hence "x \<in> \<Union> vars_rule ` ?uv ` {0 ..< n}" unfolding us_def vs_def by auto
        hence "\<gamma> x = fun_merge ?sigs (map vars_rule uvs) x" unfolding \<gamma>_def by auto
        also have "\<dots> = (?sigs ! i) x"
          by (rule fun_merge_part[OF part], insert i x ulen, auto)
        also have "\<dots> = (\<tau> i \<circ>s \<sigma> (?j i)) x" using i by simp
        finally have "\<gamma> x = (\<tau> i \<circ>s \<sigma> (?j i)) x" .
      } note \<gamma> = this
      {
        fix i
        assume i: "i < n"
        have "us i \<cdot> \<gamma> = us i \<cdot> (\<tau> i \<circ>s \<sigma> (?j i))"
          by (rule term_subst_eq, rule \<gamma>[OF i], unfold us_def vars_rule_def, auto)
        also have "\<dots> = s (?j i) \<cdot> \<sigma> (?j i)" unfolding st_uv[OF i] by simp
        finally have "us i \<cdot> \<gamma> = s (?j i) \<cdot> \<sigma> (?j i)" .
      } note u_s = this
      {
        fix i
        assume i: "i < n"
        have "vs i \<cdot> \<gamma> = vs i \<cdot> (\<tau> i \<circ>s \<sigma> (?j i))"
          by (rule term_subst_eq, rule \<gamma>[OF i], unfold vs_def vars_rule_def, auto)
        also have "\<dots> = t (?j i) \<cdot> \<sigma> (?j i)" unfolding st_uv[OF i] by simp
        finally have "vs i \<cdot> \<gamma> = t (?j i) \<cdot> \<sigma> (?j i)" .
      } note v_t = this
      have gammaF: "normal_F_subst \<gamma>" unfolding normal_F_subst_def
      proof -
        {
          fix x
          have "funas_term (\<gamma> x) \<subseteq> F \<and> \<gamma> x \<in> NF_trs R"
          proof (cases "x \<in> \<Union> vars_rule ` ?uv ` {0 ..< n}")
            case False
            hence x: "\<gamma> x = Var x" unfolding \<gamma>_def by auto
            from NF_Var_is_Fun[of "lhss R" x] wwf_var_cond[OF wwf] have "Var x \<in> NF_trs R" by force
            thus ?thesis using x by auto
          next
            case True
            then obtain i where i: "i < n" and xx: "x \<in> vars_rule (?uv i)" by auto
            from eq_rule_mod_vars_var_cond[OF eq[OF i] var_cond[OF P_Pw]] xx
            have x: "x \<in> vars_term (us i)" unfolding vars_rule_def by auto
            hence "us i \<unrhd> Var x" by auto
            hence supt: "us i \<cdot> \<gamma> \<unrhd> \<gamma> x" by auto
            from u_s[OF i] NF[of "i+j"] NFQ have "us i \<cdot> \<gamma> \<in> NF_trs R" by auto
            from NF_subterm[OF this supt] have NF: "\<gamma> x \<in> NF_trs R" .
            from \<tau>[OF i] obtain y where \<tau>: "\<tau> i x = Var y" by auto
            have "\<gamma> x = (\<tau> i \<circ>s \<sigma> (?j i)) x"
              by (rule \<gamma>[OF i], insert xx us_def vs_def, auto)
            also have "\<dots> = \<sigma> (?j i) y" unfolding subst_compose_def \<tau> by simp
            finally have "funas_term (\<gamma> x) \<subseteq> F" using \<sigma>F by auto
            with NF show ?thesis by auto
          qed
        }
        thus "\<Union>funas_term ` range \<gamma> \<subseteq> F \<and> range \<gamma> \<subseteq> NF_trs R" by auto
      qed      
      have bef: "bef < n" unfolding n_def by auto
      from cc_validE[OF valid gammaF] 
      have model: "\<gamma> \<Turnstile> constraint_of const ctype uvs bef" .
      have rewr: "Ball (set (map (\<lambda>i. CC_rewr (snd (uvs ! i)) (fst (uvs ! Suc i))) [0..<length uvs - 1])) (cc_models \<gamma>)"
      proof
        fix c
        assume "c \<in> set (map (\<lambda>i. CC_rewr (snd (uvs ! i)) (fst (uvs ! Suc i))) [0..<length uvs - 1])"
        then obtain i where i: "i < n - 1" and c: "c = CC_rewr (vs i) (us (Suc i))" unfolding us_def vs_def ulen by auto
        from i have i: "i < n" "Suc i < n" unfolding n_def by auto
        show "\<gamma> \<Turnstile> c" unfolding c cc_models.simps u_s[OF i(2)] v_t[OF i(1)] using chain 
          by (auto simp: minimal_cond_def R ichain.simps m_SN_def)
      qed
      from this model[unfolded constraint_of_def]
      have "\<gamma> \<Turnstile> condition_of const ctype (us bef, vs bef)" unfolding us_def vs_def by simp
      from tcondition_of[OF this, unfolded u_s[OF bef] v_t[OF bef]]
      have "tcondition_of ctype (s i \<cdot> \<sigma> i, t i \<cdot> \<sigma> i)" unfolding i_j_bef .
    } note main_constraint = this
    def cc \<equiv> "(Fun const []) :: ('f,'v)term"
    def u \<equiv> "\<lambda> i. s (i + bef) \<cdot> \<sigma> (i + bef)" 
    def v \<equiv> "\<lambda> i. t (i + bef) \<cdot> \<sigma> (i + bef)"
    let ?st = "\<lambda> i. (s (i + bef), t (i + bef))"
    note uv = u_def v_def
    {
      fix i
      assume i: "?st i \<in> Ps"
      from main_constraint[OF _ this cS] have "(u i, v i) \<in> S" unfolding uv by simp
    } note PS = this
    {
      fix i
      assume i: "?st i \<in> Pns"
      from main_constraint[OF _ this cNS] have "(u i, v i) \<in> NS" unfolding uv by simp
    } note PNS = this
    {
      fix i
      assume i: "?st i \<in> Pb"
      from main_constraint[OF _ this cB] have "(u i, cc) \<in> NS" unfolding uv cc_def by simp
    } note PB = this
    {
      fix i
      from ichain have steps: "(t i \<cdot> \<sigma> i, s (Suc i) \<cdot> \<sigma> (Suc i)) \<in> ?QR^*" 
       and P: "(s i, t i) \<in> P \<union> Pw" 
       and NF: "s i \<cdot> \<sigma> i \<in> NFQ" unfolding ichain.simps R by auto
      from P U_NS have orient: "gen_usable_rules (t i) \<subseteq> NS" unfolding gen_usable_rules_pairs_def by force
      from nvar[OF P] obtain f ts where ti: "t i = Fun f ts" by (cases "t i", auto)
      from PF[OF P] have "funas_args_term (t i) \<subseteq> F" .
      with \<sigma>F have tF: "funas_args_term (t i \<cdot> \<sigma> i) \<subseteq> F" unfolding ti
        by (auto simp: funas_args_term_def funas_term_subst)
      from ndef[OF P] have "?ndef (t i)" unfolding R .
      with ti have ndef: "?ndef (t i \<cdot> \<sigma> i)" by simp
      {
        fix x
        assume "x \<in> vars_term (t i)"
        with var_cond[OF P] have "x \<in> vars_term (s i)" by auto
        hence "s i \<unrhd> Var x" by auto
        from NF_subterm[OF NF supteq_stable[OF this, of "\<sigma> i"]] have "\<sigma> i x \<in> NFQ" by simp
      }
      hence NF: "\<sigma> i ` vars_term (t i) \<subseteq> NFQ" by auto
      from gen_usable_rules_lemma[OF steps orient NF tF ndef]
      have NS: "(t i \<cdot> \<sigma> i, s (Suc i) \<cdot> \<sigma> (Suc i)) \<in> NS" .
    } note NS = this
    {
      fix i
      have "(v i, u (Suc i)) \<in> NS" unfolding uv using NS[of "i+bef"] by simp
    } note NS = this
    {
      fix i
      from P_Pw[of "i + bef"] have "?st i \<in> P \<union> Pw" .
      with orient have "?st i \<in> Ps \<union> Pns" by auto
      with PS PNS have "(u i, v i) \<in> NS \<union> S" by auto
      with NS[of i]
        trans_NS[unfolded trans_def] compat have "(u i, u (Suc i)) \<in> NS \<union> S" by blast
    } note orient' = this
    {
      fix i
      assume "?st i \<in> Ps"
      from PS[OF this] NS[of i] compat have "(u i, u (Suc i)) \<in> S" by auto
    } note Ps' = this
    {
      fix i j :: nat
      assume "i \<le> j"
      hence "j = (j - i) + i" by arith
      then obtain k where j: "j = k + i" by blast
      have "(u i, u j) \<in> (NS \<union> S)^*" unfolding j
      proof (induct k)
        case 0 thus ?case by simp
      next
        case (Suc k)
        with orient'[of "k + i"] have "(u i, u (Suc k + i)) \<in> (NS \<union> S)^* O (NS \<union> S)" by auto
        thus ?case by regexp
      qed
    } note u_i_j = this
    let ?P = "\<lambda> i. ?st i \<in> Ps"
    from True have infm: "INFM i. ?P i" using Infm_shift[of "\<lambda> st. st \<in> Ps" "\<lambda> i. (s i, t i)" bef] by simp
    interpret infinitely_many ?P
      by (unfold_locales, rule infm)
    let ?i = "\<lambda> i. index i"
    def v \<equiv> "\<lambda> i. u (?i i)"
    {
      fix i
      from Ps'[OF index_p[of i]] have S: "(v i, u (Suc (?i i))) \<in> S" unfolding v_def .
      from index_ordered[of i] have "Suc (?i i) \<le> ?i (Suc i)" by simp
      from compat_rtrancl[OF S u_i_j[OF this]] have vv: "(v i, v (Suc i)) \<in> S" unfolding v_def .
      from True obtain j' where ji': "j' \<ge> ?i (Suc i) + bef" and Pb: "(s j', t j') \<in> Pb" unfolding INFM_nat_le by auto
      def j \<equiv> "j' - bef"
      from ji' Pb have ji: "j \<ge> ?i (Suc i)" and Pb: "?st j \<in> Pb" unfolding j_def by auto            
      from PB[OF Pb] have ujcc: "(u j, cc) \<in> NS" .
      from compat_rtrancl[OF vv[unfolded v_def] u_i_j[OF ji]] have "(v i, u j) \<in> S" unfolding v_def .
      with ujcc compat have "(v i, cc) \<in> S" by auto
      hence "(v i, v (Suc i)) \<in> S" and "(v i, cc) \<in> S" using vv by blast+
    } 
    with non_infE[OF non_inf, of v cc]
    show ?thesis by auto
  qed
qed

lemma conditional_general_reduction_pair_proc: 
  assumes R: "Rs \<union> Rw = R"
  and nvar: "\<And> s t. (s,t) \<in> P \<union> Pw \<Longrightarrow> is_Fun s \<and> is_Fun t"
  and var_cond: "\<And> s t. (s,t) \<in> P \<union> Pw \<Longrightarrow> vars_term t \<subseteq> vars_term s"
  and ndef: "\<And> s t. (s,t) \<in> P \<union> Pw \<Longrightarrow> \<not> defined (applicable_rules Q R) (the (root t))"
  and PF: "funas_args_trs (P \<union> Pw) \<subseteq> F"
  and U_NS: "gen_usable_rules_pairs (P \<union> Pw) \<subseteq> NS"
  and orient: "P \<union> Pw \<subseteq> Ps \<union> Pns"
  and cS: "Ball Ps (constraint_present bef aft (P \<union> Pw) Strict ccs)"
  and cNS: "Ball Pns (constraint_present bef aft (P \<union> Pw) Non_Strict ccs)"
  and cB: "Ball Pb (constraint_present bef aft (P \<union> Pw) Bound ccs)"
  and ccs: "Ball ccs cc_valid"
  and finiteS: "finite_dpp (nfs,m,P - Ps, Pw - Ps,Q,Rs,Rw)"
  and finiteB: "finite_dpp (nfs,m,P - Pb, Pw - Pb,Q,Rs,Rw)"
  shows "finite_dpp (nfs,m,P, Pw,Q,Rs,Rw)"
proof (rule ccontr)
  let ?D = "(nfs,m,P, Pw,Q,Rs,Rw)"
  let ?Ds = "(nfs,m,P - Ps, Pw - Ps,Q,Rs,Rw)"
  let ?Db = "(nfs,m,P - Pb, Pw - Pb,Q,Rs,Rw)"
  {
    from ex_inj obtain tn :: "('f,'v)term \<Rightarrow> nat" where tn: "inj tn" by auto
    from infinite_countable_subset[OF infinite_V] obtain nv :: "nat \<Rightarrow> 'v" where nv: "inj nv" by blast
    from inj_comp[OF nv tn] have "\<exists> c :: ('f,'v)term \<Rightarrow> 'v. inj c" by blast
  }
  then obtain c :: "('f,'v)term \<Rightarrow> 'v" where inj: "inj c" ..
  interpret cleaning_innermost F c by (unfold_locales, rule inj)
  assume "\<not> ?thesis"  
  then obtain s t \<sigma> where chain: "min_ichain ?D s t \<sigma>" unfolding finite_dpp_def by auto
  from R have RR: "R = Rs \<union> Rw" by simp
  from QF have QF': "QF_cond F Q" unfolding QF_cond_def by auto
  from PF have PF': "\<And> s t. (s,t) \<in> P \<union> Pw \<Longrightarrow> funas_args_term t \<subseteq> F" 
    unfolding funas_args_trs_def funas_args_rule_def by force
  from clean_min_ichain_below[of Q Rs Rw, unfolded R, OF NFQ QF' RF PF wwf _ var_cond chain]
    nvar ndef obtain \<sigma> where "min_ichain_sig ?D F s t \<sigma>" by blast
  from conditional_general_reduction_pair_chains[where Pb = Pb, OF U_NS this nvar var_cond ndef[unfolded RR] PF' orient cS cNS cB ccs R]
  have "\<exists> i. min_ichain_sig ?Ds F (shift s i) (shift t i) (shift \<sigma> i) \<or> min_ichain_sig ?Db F (shift s i) (shift t i) (shift \<sigma> i)" .
  then obtain s t \<sigma> where "min_ichain_sig ?Ds F s t \<sigma> \<or> min_ichain_sig ?Db F s t \<sigma>" by blast
  hence "min_ichain ?Ds s t \<sigma> \<or> min_ichain ?Db s t \<sigma>" unfolding min_ichain_sig.simps by blast
  with finiteS finiteB show False unfolding finite_dpp_def by blast
qed

lemma general_reduction_pair_proc: 
  assumes R: "Rs \<union> Rw = R"
  and nvar: "\<And> s t. (s,t) \<in> P \<union> Pw \<Longrightarrow> is_Fun s \<and> is_Fun t"
  and var_cond: "\<And> s t. (s,t) \<in> P \<union> Pw \<Longrightarrow> vars_term t \<subseteq> vars_term s"
  and ndef: "\<And> s t. (s,t) \<in> P \<union> Pw \<Longrightarrow> \<not> defined (applicable_rules Q R) (the (root t))"
  and PF: "funas_args_trs (P \<union> Pw) \<subseteq> F"
  and U_NS: "gen_usable_rules_pairs (P \<union> Pw) \<subseteq> NS"
  and orient: "P \<union> Pw \<subseteq> Ps \<union> Pns"
  and orientS: "Ps \<subseteq> S"
  and orientNS: "Pns \<subseteq> NS"
  and orientB: "\<And> s t. (s,t) \<in> Pb \<Longrightarrow> (s,Fun const []) \<in> NS"
  and finiteS: "finite_dpp (nfs,m,P - Ps, Pw - Ps,Q,Rs,Rw)"
  and finiteB: "finite_dpp (nfs,m,P - Pb, Pw - Pb,Q,Rs,Rw)"
  shows "finite_dpp (nfs,m,P, Pw,Q,Rs,Rw)"
proof (rule conditional_general_reduction_pair_proc[OF R nvar var_cond ndef PF U_NS orient _ _ _ _ finiteS finiteB])
  let ?c = "\<lambda> ctype st. constraint_of const ctype [st] 0"
  let ?cS = "(?c Strict) ` Ps"
  let ?cNS = "(?c Non_Strict) ` Pns"
  let ?cB = "(?c Bound) ` Pb"  
  let ?ccs = "?cS \<union> ?cNS \<union> ?cB"
  {
    fix ctype PP
    have "Ball PP (constraint_present 0 0 (P \<union> Pw) ctype (?c ctype ` PP))"
    proof (intro ballI, unfold constraint_present_def, intro allI impI)
      fix st sts
      assume mem: "st \<in> PP" and sts: "sts \<in> initial_conditions 0 0 (P \<union> Pw) st"
      obtain s t where st: "st = (s,t)" by force
      from sts[unfolded initial_conditions_def initial_conditions_gen_def Let_def] st have sts: "sts = [(s,t)]" by simp
      have disj: "disjoint_variant sts [(s,t)]" unfolding sts disjoint_variant_def
        by (auto simp: is_partition_def)
      show "\<exists>c uvs. disjoint_variant sts uvs \<and> c \<in> (?c ctype) ` PP \<and> c \<longrightarrow>cc constraint_of const ctype uvs 0"
        by (intro exI conjI, rule disj, insert st mem, auto)
    qed
  } note present = this
  from present[of Ps Strict]
  show "Ball Ps (constraint_present 0 0 (P \<union> Pw) Strict ?ccs)" by (force simp: constraint_present_def)
  from present[of Pns Non_Strict]
  show "Ball Pns (constraint_present 0 0 (P \<union> Pw) Non_Strict ?ccs)" by (force simp: constraint_present_def)
  from present[of Pb Bound]
  show "Ball Pb (constraint_present 0 0 (P \<union> Pw) Bound ?ccs)" by (force simp: constraint_present_def)
  show "Ball ?ccs cc_valid"
  proof
    fix cc
    assume cc: "cc \<in> ?ccs"
    show "cc_valid cc"
    proof
      fix \<sigma>
      assume "normal_F_subst \<sigma>"
      hence \<sigma>F: "\<Union> funas_term ` range \<sigma> \<subseteq> F" unfolding normal_F_subst_def by auto
      note stable_S = F_subst_closedE[OF stable_S \<sigma>F]
      note stable_NS = F_subst_closedE[OF stable_NS \<sigma>F]
      {
        assume "cc \<in> ?cS"
        then obtain s t where st: "(s,t) \<in> Ps" and cc: "cc = ?c Strict (s,t)" by force
        from st orientS have "(s,t) \<in> S" by auto
        from stable_S[OF this]
        have "\<sigma> \<Turnstile> cc" unfolding cc constraint_of_def by (simp add: rel_of_def)
      }
      moreover
      {
        assume "cc \<in> ?cB"
        then obtain s t where st: "(s,t) \<in> Pb" and cc: "cc = ?c Bound (s,t)" by force
        from st orientB have "(s,Fun const []) \<in> NS" by auto
        from stable_NS[OF this]
        have "\<sigma> \<Turnstile> cc" unfolding cc constraint_of_def by (simp add: rel_of_def)
      }
      moreover
      {
        assume "cc \<in> ?cNS"
        then obtain s t where st: "(s,t) \<in> Pns" and cc: "cc = ?c Non_Strict (s,t)" by force
        from st orientNS have "(s,t) \<in> NS" by auto
        from stable_NS[OF this]
        have "\<sigma> \<Turnstile> cc" unfolding cc constraint_of_def by (simp add: rel_of_def)
      }
      ultimately show "\<sigma> \<Turnstile> cc" using cc by blast
    qed
  qed
qed auto
end

definition distance :: "'a rel \<Rightarrow> 'a set \<Rightarrow> 'a \<Rightarrow> nat" where
  "distance r s a \<equiv> LEAST n. \<exists> b \<in> s. (a,b) \<in> r ^^ n"

lemma distance_path: assumes b: "b \<in> s" and path: "(a,b) \<in> r^*"
  shows "\<exists> b \<in> s. (a,b) \<in> r ^^ distance r s a"
proof -
  let ?p = "\<lambda> n. \<exists> b \<in> s. (a,b) \<in> r ^^ n"
  from rtrancl_imp_relpow[OF path] b have "\<exists> n. ?p n" by blast
  from LeastI_ex[of ?p, OF this]
  show ?thesis unfolding distance_def .
qed

lemma distance_least: assumes "b \<in> s" and "(a,b) \<in> r^^n"
  shows "distance r s a \<le> n"
unfolding distance_def
  by (rule Least_le, insert assms, auto)

lemma distance_Suc_relpow: assumes b: "b \<in> s" and path: "(a,b) \<in> r ^^ distance r s a" and a: "a \<noteq> b"
  shows "\<exists> c. (a,c) \<in> r \<and> (c,b) \<in> r ^^ distance r s c \<and> distance r s a = Suc (distance r s c)"
proof -
  from a path obtain n where d: "distance r s a = Suc n" by (cases "distance r s a", auto)
  from relpow_Suc_D2[OF path[unfolded d]] obtain c where ac: "(a,c) \<in> r" and cb: "(c,b) \<in> r ^^ n" by auto
  from distance_least[OF b cb] have n: "distance r s c \<le> n" .
  {
    assume n: "distance r s c < n"
    from distance_path[OF b relpow_imp_rtrancl[OF cb]] obtain b' where b': "b' \<in> s" and cb': "(c,b') \<in> r ^^ distance r s c" ..
    from distance_least[OF b' relpow_Suc_I2[OF ac cb'], unfolded d] n have False by arith
  }
  with n have n: "n = distance r s c" by arith
  show ?thesis
    by (intro exI conjI, rule ac, insert cb d n, auto)
qed

lemma distance_Suc_rtrancl: assumes b: "b \<in> s" and path: "(a,b) \<in> r^*" and a: "a \<notin> s"
  shows "\<exists> c b. (a,c) \<in> r \<and> (c,b) \<in> r^* \<and> distance r s a = Suc (distance r s c)"
proof -
  from distance_path[OF b path] obtain b' where b': "b' \<in> s" "(a,b') \<in> r ^^ distance r s a" by blast
  from a b' have "a \<noteq> b'" by auto
  from distance_Suc_relpow[OF b' this] obtain c where ac: "(a, c) \<in> r" 
    and cb: "(c, b') \<in> r ^^ distance r s c" and d: "distance r s a = Suc (distance r s c)" by blast
  from ac relpow_imp_rtrancl[OF cb] d show ?thesis by blast
qed

lemma distance_subterm_innermost: fixes Q R nfs
  defines dist: "dist \<equiv> distance (qrstep nfs Q R) (NF_terms Q)"
  assumes b: "b \<in> NF_terms Q" and ab: "(a,b) \<in> (qrstep nfs Q R)^*"
  and ac: "a \<unrhd> c"
  shows "dist a \<ge> dist c"
proof -
  from distance_path[OF b ab] obtain b where
    b: "b \<in> NF_terms Q" and ab: "(a, b) \<in> qrstep nfs Q R ^^ dist a" unfolding dist by auto
  from supteq_imp_subt_at[OF ac] obtain p where p: "p \<in> poss a" and c: "c = a |_ p" by auto
  from normalize_subterm_qrsteps_count[OF p ab b]
    obtain n1 n2 u where cu: "(c, u) \<in> qrstep nfs Q R ^^ n1" and u: "u \<in> NF_terms Q" and da: "dist a = n1 + n2"
    unfolding c by blast
  from distance_least[OF u cu] have "dist c \<le> n1" unfolding dist by auto
  also have "\<dots> \<le> dist a" unfolding da by simp
  finally show ?thesis .
qed

definition UNF_r_q where 
  "UNF_r_q r q \<equiv> \<forall> a b c. (a,b) \<in> r^* \<longrightarrow> (a,c) \<in> r^* \<longrightarrow> b \<in> q \<longrightarrow> c \<in> q \<longrightarrow> b = c"

lemma CR_imp_UNF_inn: assumes inn: "NF_terms Q \<subseteq> NF_trs R" and sub: "R' \<subseteq> R" and cr: "CR (rstep R')" 
  shows "UNF_r_q (qrstep nfs Q R') (NF_terms Q)"
proof -
  let ?QR = "qrstep nfs Q R"
  let ?QR' = "qrstep nfs Q R'"
  let ?Q = "NF_terms Q"
  let ?R' = "rstep R'"
  show ?thesis 
  unfolding UNF_r_q_def
  proof (intro allI impI)
    fix a b c
    assume ab: "(a,b) \<in> ?QR'^*" and ac: "(a,c) \<in> ?QR'^*" and b: "b \<in> ?Q" and c: "c \<in> ?Q"
    from inn NF_anti_mono[OF rstep_mono[OF sub]] have inn: "?Q \<subseteq> NF ?R'" by auto
    have sub: "?QR'^* \<subseteq> ?R'^*" by (rule rtrancl_mono, auto)
    from sub ab ac have ab: "(a,b) \<in> ?R'^*" and ac: "(a,c) \<in> ?R'^*" by auto
    from inn b c have b: "b \<in> NF ?R'" and c: "c \<in> NF ?R'" by auto
    from CR_imp_UNF[OF cr] ab ac b c show "b = c" by auto
  qed
qed    
  

lemma distance_NF_UNF: assumes NF: "s \<subseteq> q" and UNF: "UNF_r_q r q" and b: "b \<in> s" and path: "(a,b) \<in> r^*"
  shows "(a,b) \<in> r^^distance r s a"
proof -
  from distance_path[OF b path] obtain b' where b': "b' \<in> s" and ab': "(a,b') \<in> r ^^ distance r s a" ..
  from b b' NF have "b \<in> q" "b' \<in> q" by auto
  with path relpow_imp_rtrancl[OF ab'] UNF[unfolded UNF_r_q_def] have "b' = b" by auto
  with ab' show ?thesis by auto
qed  

locale fresh =
  fixes fresh :: "'v list \<Rightarrow> 'v"
begin

fun cc_subst_apply :: "('f,'v)cond_constraint \<Rightarrow> (('f,'v)subst \<times> 'v list) \<Rightarrow> ('f,'v)cond_constraint" (infix "\<cdot>cc" 70) where 
  "CC_cond ct (s,t) \<cdot>cc (\<sigma>,_) = CC_cond ct (s \<cdot> \<sigma>,t \<cdot> \<sigma>)"
| "CC_rewr s t \<cdot>cc (\<sigma>,_) = CC_rewr (s \<cdot> \<sigma>) (t \<cdot> \<sigma>)"
| "CC_impl c1 c2 \<cdot>cc \<sigma> = CC_impl (map (\<lambda> c. c \<cdot>cc \<sigma>) c1) (c2 \<cdot>cc \<sigma>)"
| "CC_all x c \<cdot>cc (\<sigma>,vs) = (let y = fresh (vs @ vars_cc_list (CC_all x c)) in 
     CC_all y (c \<cdot>cc (\<sigma>(x := Var y), y # vs)))"

definition normalize_alpha :: "('f,'v)cond_constraint \<Rightarrow> ('f,'v)cond_constraint" where
  "normalize_alpha c \<equiv> c \<cdot>cc (Var,[])"

definition cc_ih_prems :: "'f \<Rightarrow> ('f,'v)term \<Rightarrow> 'v list 
  \<Rightarrow> ('f,'v)cond_constraint list \<Rightarrow> ('f,'v)cond_constraint \<Rightarrow> (('f,'v)term \<times> 'v list) list \<Rightarrow> ('f,'v)cond_constraint list" where 
  "cc_ih_prems f q xs \<phi> \<psi> rs_ys_list \<equiv> 
      map (\<lambda> (r,ys). let 
            rs = args r;
            \<mu> = mk_subst Var (zip xs rs);
            vs' = vars_subst_range_impl (zip xs rs);
            mu = \<lambda> c. c \<cdot>cc (\<mu>,vs');
            \<phi>' = CC_impl (CC_rewr r (q \<cdot> \<mu>) # map mu \<phi>) (mu \<psi>)
           in cc_bound ys \<phi>') rs_ys_list"

definition cc_rule_constraint :: "'f \<Rightarrow> ('f,'v)term list \<Rightarrow> ('f,'v)term \<Rightarrow> ('f,'v)term \<Rightarrow> 'v list 
  \<Rightarrow> ('f,'v)cond_constraint list \<Rightarrow> ('f,'v)cond_constraint \<Rightarrow> (('f,'v)term \<times> 'v list) list \<Rightarrow> ('f,'v)cond_constraint" where 
  "cc_rule_constraint f ls r q xs \<phi> \<psi> rs_ys_list \<equiv> 
     let 
       \<sigma> = mk_subst Var (zip xs ls);
       vs = vars_subst_range_impl (zip xs ls);
       rew = CC_rewr r (q \<cdot> \<sigma>);
       phi_sig = map (\<lambda>c. c \<cdot>cc (\<sigma>,vs)) \<phi>;
       psi_sig = \<psi> \<cdot>cc (\<sigma>,vs);
       ihs = cc_ih_prems f q xs \<phi> \<psi> rs_ys_list
     in CC_impl (rew # phi_sig @ ihs) psi_sig"

end

declare fresh.cc_subst_apply.simps[code]
declare fresh.normalize_alpha_def[code]
declare fresh.cc_rule_constraint_def[code]
declare fresh.cc_ih_prems_def[code]

locale fixed_trs_dep_order_fresh = fixed_trs_dep_order \<pi> R Q F S NS nfs m const + fresh fresh
  for \<pi> :: "('f :: countable) dep"
  and R :: "('f,'v :: countable)trs"
  and Q :: "('f,'v)term set"
  and F :: "'f sig"
  and S :: "('f,'v)trs"
  and NS :: "('f,'v)trs"
  and nfs :: bool
  and m :: bool
  and fresh :: "'v list \<Rightarrow> 'v"
  and const :: 'f
  + fixes UR :: "('f,'v)term \<Rightarrow> ('f,'v)trs"
  assumes fresh: "fresh vs \<notin> set vs"
  and UR: "\<And> t u \<sigma>. \<sigma> ` vars_term t \<subseteq> NF_terms Q \<Longrightarrow> (t \<cdot> \<sigma>, u) \<in> (qrstep nfs Q R)^* \<Longrightarrow> (t \<cdot> \<sigma>, u) \<in> (qrstep nfs Q (UR t))^*"
  and UR_sub: "\<And> t. UR t \<subseteq> R"
begin


lemma cc_models_vars: assumes "\<And> x. x \<in> vars_cc c \<Longrightarrow> \<sigma> x = \<tau> x"
  shows "\<sigma> \<Turnstile> c = \<tau> \<Turnstile> c"
  using assms
proof (induct c arbitrary: \<sigma> \<tau>)
  case (cond ct s t)
  have "(Fun f [s,t,Fun const []]) \<cdot> \<sigma> = (Fun f [s,t,Fun const []] \<cdot> \<tau>)"
    by (rule term_subst_eq, insert cond, auto)
  thus ?case by auto
next
  case (rewr s t)
  have "(Fun f [s,t]) \<cdot> \<sigma> = (Fun f [s,t] \<cdot> \<tau>)"
    by (rule term_subst_eq, insert rewr, auto)
  thus ?case by auto
next
  case (impl c1 c2 \<sigma> \<tau>)
  from this(1)[of _ \<sigma> \<tau>] this(2)[of \<sigma> \<tau>] this(3)
  show ?case by auto
next
  case (all x c \<sigma> \<tau>)
  {
    fix t
    have "\<sigma>(x := t) \<Turnstile> c = \<tau>(x := t) \<Turnstile> c"
      by (rule all(1), insert all(2), auto)
  }
  thus ?case by auto
qed

lemma cc_models_subst: assumes vs: "vars_subst_range \<sigma>1 \<subseteq> set vs"
  shows "(\<sigma>2 \<Turnstile> c \<cdot>cc (\<sigma>1,vs)) = (\<sigma>1 \<circ>s \<sigma>2 \<Turnstile> c)"
  using vs
proof (induct c arbitrary: \<sigma>1 \<sigma>2 vs)
  case (all x c \<sigma>1 \<sigma>2 vs)
  let ?Var = "Var :: 'v \<Rightarrow> ('f,'v)term"
  obtain y where y: "fresh (vs @ vars_cc_list (CC_all x c)) = y" by auto
  with fresh[of "vs @ vars_cc_list (CC_all x c)"] have fresh: "y \<notin> set vs \<union> vars_cc (CC_all x c)" by auto
  with all(2) have y1: "y \<notin> vars_subst_range \<sigma>1" by auto
  let ?vs = "y # vs"
  have "vars_subst_range (\<sigma>1(x := ?Var y)) \<subseteq> set ?vs"
  proof 
    fix z
    assume "z \<in> vars_subst_range (\<sigma>1(x := ?Var y))"
    then obtain u where u: "u \<in> subst_domain (\<sigma>1(x := ?Var y))" and z: "z \<in> vars_term ((\<sigma>1(x := ?Var y)) u)"
      unfolding vars_subst_range_def by auto
    show "z \<in> set ?vs"
    proof (cases "u = x")
      case True
      with z show ?thesis by auto
    next
      case False
      with u have "u \<in> subst_domain \<sigma>1" unfolding subst_domain_def by auto
      with z False have "z \<in> vars_subst_range \<sigma>1" unfolding vars_subst_range_def by auto
      with all(2) show ?thesis by auto
    qed
  qed
  note IH = all(1)[OF this]
  show ?case unfolding cc_subst_apply.simps y Let_def 
  proof -
    show "(\<sigma>2 \<Turnstile> CC_all y (c \<cdot>cc (\<sigma>1(x := Var y), ?vs))) = 
       \<sigma>1 \<circ>s \<sigma>2 \<Turnstile> CC_all x c" (is "?l = ?r") 
    proof -
      def P \<equiv> "\<lambda> t. funas_term t \<subseteq> F \<and> t \<in> NF_trs R"
      have l: "?l = (\<forall> t. P t \<longrightarrow> \<sigma>2( y := t) \<Turnstile> (c \<cdot>cc (\<sigma>1(x := Var y), ?vs)))" unfolding P_def by auto
      have r: "?r = (\<forall> t. P t \<longrightarrow> (\<sigma>1 \<circ>s \<sigma>2) (x := t) \<Turnstile> c)" unfolding P_def by auto
      {
        fix t
        assume "P t"
        let ?l = "(\<sigma>1(x := Var y) \<circ>s \<sigma>2(y := t))"
        let ?r = "((\<sigma>1 \<circ>s \<sigma>2)(x := t))"
        have "((\<sigma>2( y := t) \<Turnstile> (c \<cdot>cc (\<sigma>1(x := Var y), ?vs)))
           =  ((\<sigma>1 \<circ>s \<sigma>2) (x := t) \<Turnstile> c)) = (?l \<Turnstile> c = ?r \<Turnstile> c)"
          unfolding IH by simp
        also have "(?l \<Turnstile> c = ?r \<Turnstile> c) = True"
        proof -
          {
            fix z
            assume z: "z \<in> vars_cc c"
            have "?l z = ?r z"
            proof (cases "z = x")
              case True
              thus ?thesis by (simp add: subst_compose_def)
            next
              case False
              with z fresh have neq: "y \<noteq> z" by auto
              have "(?l z = ?r z) = (\<sigma>1 z \<cdot> \<sigma>2(y := t) =  \<sigma>1 z \<cdot> \<sigma>2)" unfolding subst_compose_def using False by simp
              moreover have "..."
              proof (cases "z \<in> subst_domain \<sigma>1")
                case True
                hence "vars_term (\<sigma>1 z) \<subseteq> vars_subst_range \<sigma>1" unfolding vars_subst_range_def by auto
                with y1 have y: "y \<notin> vars_term (\<sigma>1 z)" by auto
                show ?thesis
                  by (rule term_subst_eq, insert y, auto)
              next
                case False
                hence z1: "\<sigma>1 z = Var z" unfolding subst_domain_def by auto
                show ?thesis unfolding z1 using neq by simp
              qed
              ultimately show ?thesis by simp
            qed
          } note main = this
          hence "(?l \<Turnstile> c = ?r \<Turnstile> c)" by (rule cc_models_vars)
          thus ?thesis by simp
        qed
        finally have "(\<sigma>2( y := t) \<Turnstile> (c \<cdot>cc (\<sigma>1(x := Var y), ?vs)))
           =  ((\<sigma>1 \<circ>s \<sigma>2) (x := t) \<Turnstile> c)" by simp
      } thus ?thesis unfolding l r by auto
    qed
  qed
qed auto

lemma normalize_alpha: "\<sigma> \<Turnstile> normalize_alpha c = \<sigma> \<Turnstile> c"
proof -
  have id: "\<sigma> \<Turnstile> c = Var \<circ>s \<sigma> \<Turnstile> c" by simp
  show ?thesis unfolding normalize_alpha_def id 
    by (rule cc_models_subst, unfold vars_subst_range_def, auto)
qed

lemma alpha_equivalence: assumes eq: "normalize_alpha c = normalize_alpha d"
  shows "c =cc d"
proof (rule cc_equivI)
  fix \<sigma>
  have "\<sigma> \<Turnstile> c = \<sigma> \<Turnstile> normalize_alpha c" unfolding normalize_alpha ..
  also have "... = \<sigma> \<Turnstile> normalize_alpha d" unfolding eq ..
  also have "... = \<sigma> \<Turnstile> d" unfolding normalize_alpha ..
  finally show "\<sigma> \<Turnstile> c = \<sigma> \<Turnstile> d" .
qed

lemma cc_models_boundE: assumes ys: "\<And> x. x \<notin> set ys \<Longrightarrow> \<sigma> x = \<delta> x"
  and models: "\<sigma> \<Turnstile> cc_bound ys c"
  and F: "normal_F_subst \<delta>"
  shows "\<delta> \<Turnstile> c"
  using ys models
proof (induct ys arbitrary: \<sigma>)
  case Nil
  thus ?case by simp
next
  case (Cons y ys \<sigma>)
  let ?y = "\<delta> y"
  from F[unfolded normal_F_subst_def] have "funas_term ?y \<subseteq> F" and "?y \<in> NF_trs R" by auto
  with Cons(3) have model: "\<sigma>(y := ?y) \<Turnstile> cc_bound ys c" by auto
  show ?case
    by (rule Cons(1)[OF _ model], insert Cons(2), auto)
qed

lemma cc_models_boundI: 
  assumes ys: "\<And> \<sigma>. \<lbrakk>\<And>x. x \<notin> set ys \<Longrightarrow> \<sigma> x = \<delta> x\<rbrakk> \<Longrightarrow> \<lbrakk>\<And>x. x \<in> set ys \<Longrightarrow> funas_term (\<sigma> x) \<subseteq> F \<and> \<sigma> x \<in> NF_trs R\<rbrakk> \<Longrightarrow> \<sigma> \<Turnstile> c"
  shows "\<delta> \<Turnstile> cc_bound ys c"
  using ys
proof (induct ys arbitrary: \<delta>)
  case Nil
  thus ?case by simp
next
  case (Cons y ys \<delta>)
  show ?case unfolding cc_bound.simps cc_models.simps
  proof (intro allI impI)
    fix t
    assume tF: "funas_term t \<subseteq> F" and t: "t \<in> NF_trs R"
    show "\<delta>(y := t) \<Turnstile> cc_bound ys c"
    proof (rule Cons(1))
      fix \<sigma>
      assume eq: "\<And>x. x \<notin> set ys \<Longrightarrow> \<sigma> x = (\<delta>(y := t)) x"
      assume fun_NF: "\<And>x. x \<in> set ys \<Longrightarrow> funas_term (\<sigma> x) \<subseteq> F \<and> \<sigma> x \<in> NF_trs R"
      show "\<sigma> \<Turnstile> c"
      proof (rule Cons(2))
        fix x
        assume "x \<notin> set (y # ys)" with eq[of x] show "\<sigma> x = \<delta> x" by auto
      next
        fix x
        assume "x \<in> set (y # ys)"
        thus "funas_term (\<sigma> x) \<subseteq> F \<and> \<sigma> x \<in> NF_trs R" using eq[of x] fun_NF[of x] tF t
          by (cases "x \<in> set ys", auto)
      qed
    qed
  qed
qed


inductive cc_simplify :: "('f,'v)cond_constraint \<Rightarrow> ('f,'v)cond_constraint list \<Rightarrow> bool" (infix "\<turnstile>" 51) where
  delete_conditions: "normalize_cc c = CC_impl \<phi> \<psi> \<Longrightarrow> normalize_cc c' = CC_impl \<phi>' \<psi> \<Longrightarrow> set \<phi>' \<subseteq> set \<phi> \<Longrightarrow> c \<turnstile> [c']"
| constructor_different: "normalize_cc c = CC_impl \<phi> \<psi> \<Longrightarrow> (f,length ss) \<noteq> (g,length ts) \<Longrightarrow> \<not> (defined R (f,length ss)) 
  \<Longrightarrow> CC_rewr (Fun f ss) (Fun g ts) \<in> set \<phi> \<Longrightarrow> c \<turnstile> []"
| constructor_same: "normalize_cc c = CC_impl \<phi> \<psi> 
  \<Longrightarrow> c' \<longrightarrow>cc CC_impl (\<phi> @ map (\<lambda> (s,t). CC_rewr s t) (zip ss ts)) \<psi> 
  \<Longrightarrow> length ss = length ts \<Longrightarrow> \<not> (defined R (f,length ss)) 
  \<Longrightarrow> CC_rewr (Fun f ss) (Fun f ts) \<in> set \<phi>  
  \<Longrightarrow> c \<turnstile> [c']" 
| variable_left: "normalize_cc c = CC_impl \<phi> \<psi> 
  \<Longrightarrow> c' \<longrightarrow>cc c \<cdot>cc \<tau>
  \<Longrightarrow> \<tau> = (Var(x := q),vars_term_list q) \<Longrightarrow> CC_rewr (Var x) q \<in> set \<phi>
  \<Longrightarrow> c \<turnstile> [c']"
| variable_right: "normalize_cc c = CC_impl \<phi> \<psi>  
  \<Longrightarrow> c' \<longrightarrow>cc c \<cdot>cc \<tau>
  \<Longrightarrow> \<tau> = (Var(x := q),vars_term_list q) \<Longrightarrow> CC_rewr q (Var x) \<in> set \<phi>
  \<Longrightarrow> \<lbrakk> \<And> fn. fn \<in> funas_term q \<Longrightarrow> \<not> defined R fn\<rbrakk> 
  \<Longrightarrow> c \<turnstile> [c']"
| simplify_condition: "normalize_cc c = CC_impl \<phi> \<psi>
  \<Longrightarrow> cc_bound ys cc \<in> set \<phi>
  \<Longrightarrow> normalize_cc cc = CC_impl \<phi>' \<psi>'
  \<Longrightarrow> subst_domain \<sigma> \<subseteq> set ys 
  \<Longrightarrow> vars_subst_range \<sigma> \<subseteq> set vs 
  \<Longrightarrow> \<lbrakk>\<And> fn. fn \<in> \<Union> funas_term ` subst_range \<sigma> \<Longrightarrow> fn \<in> F \<and> \<not> (defined R fn)\<rbrakk> 
  \<Longrightarrow> \<lbrakk>\<And> c. c \<in> set \<phi>' \<Longrightarrow> \<exists> d \<in> set \<phi>. d \<longrightarrow>cc c \<cdot>cc (\<sigma>,vs)\<rbrakk> 
  \<Longrightarrow> c' \<longrightarrow>cc CC_impl (\<psi>' \<cdot>cc (\<sigma>,vs) # \<phi>) \<psi>
  \<Longrightarrow> c \<turnstile> [c']"
| fun_arg_into_var: "normalize_cc c = CC_impl \<phi> \<psi> 
  \<Longrightarrow> c' \<longrightarrow>cc CC_impl (CC_rewr p (Var x) # CC_rewr (Fun f (bef @ Var x # aft)) q # \<phi>) \<psi>
  \<Longrightarrow> x \<notin> vars_cc c 
  \<Longrightarrow> CC_rewr (Fun f (bef @ p # aft)) q \<in> set \<phi> 
  \<Longrightarrow> funas_term p \<subseteq> F
  \<Longrightarrow> c \<turnstile> [c']"
| induction: "normalize_cc c = CC_impl \<phi>' \<psi>  
  \<Longrightarrow> mgu (Fun f (map Var xs)) q = None
  \<Longrightarrow> CC_rewr (Fun f (map Var xs)) q \<in> set \<phi>'
  \<Longrightarrow> set \<phi> \<subseteq> set \<phi>'
  \<Longrightarrow> \<lbrakk> \<And> ls r. length xs = length ls \<Longrightarrow> (Fun f ls,r) \<in> R \<Longrightarrow> \<exists> d ls' r' r_ys_list. 
    (Fun f ls,r) =v (Fun f ls',r') \<and> vars_rule (Fun f ls',r') \<inter> vars_cc (CC_impl (CC_rewr (Fun f (map Var xs)) q # \<phi>) \<psi>) = {} \<and> d \<in> set cs \<and> d \<longrightarrow>cc cc_rule_constraint f ls' r' q xs \<phi> \<psi> r_ys_list
    \<and> (\<forall>(r, ys) \<in> set r_ys_list. root r = Some (f,length xs) \<and> r' \<unrhd> r \<and> vars_term r' \<inter> set ys = {} \<and> (\<forall> fn \<in> funas_args_term r. \<not> defined R fn))\<rbrakk> 
  \<Longrightarrow> distinct xs
  \<Longrightarrow> m \<or> CR (rstep (UR (Fun f (map Var xs))))
  \<Longrightarrow> c \<turnstile> cs"

lemma normalize_validI: "normalize_cc c = CC_impl \<phi> \<psi> \<Longrightarrow> 
  \<lbrakk>\<And> \<sigma>. normal_F_subst \<sigma> \<Longrightarrow> \<lbrakk> \<And> c'. c' \<in> set \<phi> \<Longrightarrow> \<sigma> \<Turnstile> c' \<rbrakk> \<Longrightarrow> \<sigma> \<Turnstile> \<psi>\<rbrakk> \<Longrightarrow> cc_valid c" 
  using normalize_cc[of _ c] unfolding cc_valid_def by auto

lemma normalize_validE: "normalize_cc c = CC_impl \<phi> \<psi> \<Longrightarrow> cc_valid c \<Longrightarrow> normal_F_subst \<sigma> \<Longrightarrow>
  \<lbrakk> \<And> c'. c' \<in> set \<phi> \<Longrightarrow> \<sigma> \<Turnstile> c'\<rbrakk> \<Longrightarrow> \<sigma> \<Turnstile> \<psi>"
  using normalize_cc[of _ c] unfolding cc_valid_def by auto

lemma cc_simplify: "c \<turnstile> c' \<Longrightarrow> Ball (set c') cc_valid \<Longrightarrow> cc_valid c"
proof -
  from wwf_var_cond[OF wwf] have Rfun: "\<forall>(l, r)\<in>R. is_Fun l" .
  note a_ndef = ndef_applicable_rules[of R _ Q]
  {
    fix \<delta> :: "('f,'v)subst"
    fix vs :: "'v list"
    fix x :: 'v
    fix q :: "('f,'v)term"
    fix \<phi> :: "('f,'v)cond_constraint"
    assume vs: "vars_term q \<subseteq> set vs"
    assume id: "\<delta> x = q \<cdot> \<delta>"
    have "(\<delta> \<Turnstile> \<phi> \<cdot>cc (Var(x := q),vs)) = (Var(x := q) \<circ>s \<delta> \<Turnstile> \<phi>)"
    proof (rule cc_models_subst)
      show "vars_subst_range (Var(x := q)) \<subseteq> set vs" unfolding vars_subst_range_def subst_range.simps subst_domain_def using vs
        by auto
    qed
    also have "(Var(x := q) \<circ>s \<delta>) = \<delta>"
    proof
      fix y
      show "(Var(x := q) \<circ>s \<delta>) y = \<delta> y"
        by (unfold subst_compose_def, cases "y = x", insert id, auto)
    qed
    finally have "(\<delta> \<Turnstile> \<phi> \<cdot>cc (Var(x := q),vs)) = (\<delta> \<Turnstile> \<phi>)" .
  } note subst = this
  show "c \<turnstile> c' \<Longrightarrow> Ball (set c') cc_valid \<Longrightarrow> cc_valid c"
  proof (induct rule: cc_simplify.induct)
    case (delete_conditions c \<phi> \<psi> c' \<chi>)
    hence c': "cc_valid c'" by auto
    show ?case
      by (rule normalize_validI[OF delete_conditions(1)],
        insert normalize_validE[OF delete_conditions(2) c'] delete_conditions(3), auto)
  next
    case (constructor_different c \<phi> \<psi> f ss g ts)
    show ?case
    proof (rule normalize_validI[OF constructor_different(1)])
      fix \<sigma>
      assume phi: "\<And> c'. c' \<in> set \<phi> \<Longrightarrow> \<sigma> \<Turnstile> c'"
      have "(Fun f ss \<cdot> \<sigma>, Fun g ts \<cdot> \<sigma>) \<notin> (qrstep nfs Q R)^*"
      proof
        let ?\<sigma> = "map (\<lambda> t. t \<cdot> \<sigma>)"
        from constructor_different(3) have ndef: "\<not> defined R (the (root (Fun f (?\<sigma> ss))))" by simp
        assume "(Fun f ss \<cdot> \<sigma>, Fun g ts \<cdot> \<sigma>) \<in> (qrstep nfs Q R)^*"
        hence "(Fun f (?\<sigma> ss), Fun g (?\<sigma> ts)) \<in> (qrstep nfs Q R)^*" by simp
        from nrqrsteps_preserve_root[OF qrsteps_imp_nrqrsteps[OF Rfun a_ndef[OF ndef] this]]
        show False using constructor_different(2) by simp
      qed
      with phi[OF constructor_different(4)]
      show "\<sigma> \<Turnstile> \<psi>" by auto
    qed
  next
    case (constructor_same c \<phi> \<psi> c' ss ts f)
    hence c': "cc_valid c'" by auto
    note valid = cc_validE[OF cc_impliesE[OF constructor_same(2) c']]
    show ?case
    proof (rule normalize_validI[OF constructor_same(1)])
      fix \<sigma>
      assume nfs: "normal_F_subst \<sigma>"
      note valid = valid[OF nfs]
      assume phi: "\<And>c'. c' \<in> set \<phi> \<Longrightarrow> \<sigma> \<Turnstile> c'"
      from phi[OF constructor_same(5)] have
        mrewr: "\<sigma> \<Turnstile> CC_rewr (Fun f ss) (Fun f ts)" by auto
      from constructor_same have len: "length ss = length ts" by simp
      have "Ball (set (map (\<lambda>(s, t). CC_rewr s t) (zip ss ts))) (cc_models \<sigma>) = 
        (\<forall> i. i < length ss \<longrightarrow> \<sigma> \<Turnstile> CC_rewr (ss ! i) (ts ! i))" using len
        by (auto simp: set_zip)
      also have "..."
      proof (intro allI impI)
        fix i
        assume i: "i < length ss"
        let ?\<sigma> = "map (\<lambda> t. t \<cdot> \<sigma>)"
        from a_ndef[OF constructor_same(4)] have ndef: "\<not> defined (applicable_rules Q R) (the (root (Fun f (?\<sigma> ss))))" by simp
        from mrewr have steps: "(Fun f (?\<sigma> ss), Fun f (?\<sigma> ts)) \<in> (qrstep nfs Q R)^*" by simp
        from mrewr have nf: "Fun f (?\<sigma> ts) \<in> NF_terms Q" by simp
        from nrqrsteps_imp_arg_qrsteps[OF qrsteps_imp_nrqrsteps[OF Rfun ndef steps], of i] i len
        have steps: "(ss ! i \<cdot> \<sigma>, ts ! i \<cdot> \<sigma>) \<in> (qrstep nfs Q R)^*" by auto
        have nf: "ts ! i \<cdot> \<sigma> \<in> NF_terms Q" by (rule NF_subterm[OF nf], insert i len, auto)
        {
          assume m
          with mrewr have SN: "SN_on (qrstep nfs Q R) {Fun f (?\<sigma> ss)}" unfolding m_SN_def cc_models.simps by simp
          have SN: "SN_on (qrstep nfs Q R) {ss ! i \<cdot> \<sigma>}"
            by (rule SN_imp_SN_arg_gen[OF ctxt_closed_qrstep SN], insert i, auto)
        }
        hence SN: "m_SN (ss ! i \<cdot> \<sigma>)" unfolding m_SN_def by simp
        show "\<sigma> \<Turnstile> CC_rewr (ss ! i) (ts ! i)" using steps nf SN by auto
      qed
      finally show "\<sigma> \<Turnstile> \<psi>" using valid phi by auto
    qed
  next
    case (variable_left c \<phi> \<psi> c' \<tau> x q)
    hence c': "cc_valid c'" by auto
    note valid = cc_validE[OF cc_impliesE[OF variable_left(2) c']]
    show ?case
    proof (rule normalize_validI[OF variable_left(1)])
      fix \<sigma>
      assume sigma: "normal_F_subst \<sigma>"
      assume phi: "\<And>c'. c' \<in> set \<phi> \<Longrightarrow> \<sigma> \<Turnstile> c'"
      from phi[OF variable_left(4)] have 
        steps: "(\<sigma> x, q \<cdot> \<sigma>) \<in> (qrstep nfs Q R)^*" by auto
      from steps rtrancl_mono[OF qrstep_mono[of R R Q "{}" nfs]] have steps: "(\<sigma> x, q \<cdot> \<sigma>) \<in> (rstep R)^*" by auto
      from sigma[unfolded normal_F_subst_def] NFQ have sigmax: "\<sigma> x \<in> NF_trs R" by force
      from steps sigmax have id: "\<sigma> x = q \<cdot> \<sigma>" unfolding NF_def by (induct, auto)
      have subset: "vars_term q \<subseteq> set (vars_term_list q)" by auto
      note subst = subst[OF subset id, unfolded variable_left(3)[symmetric]]
      from valid[OF sigma, unfolded subst normalize_cc[of _ c] variable_left(1)] phi
      show "\<sigma> \<Turnstile> \<psi>" by auto
    qed
  next
    case (variable_right c \<phi> \<psi>  c' \<tau> x q)
    hence c': "cc_valid c'" by auto
    note valid = cc_validE[OF cc_impliesE[OF variable_right(2) c']]
    show ?case
    proof (rule normalize_validI[OF variable_right(1)])
      fix \<sigma>
      assume sigma: "normal_F_subst \<sigma>"
      assume phi: "\<And>c'. c' \<in> set \<phi> \<Longrightarrow> \<sigma> \<Turnstile> c'"
      from phi[OF variable_right(4)] have 
        steps: "(q \<cdot> \<sigma>, \<sigma> x) \<in> (qrstep nfs Q R)^*" by auto
      have "q \<cdot> \<sigma> \<in> NF (qrstep nfs Q R)"
      proof (rule NF_subst_qrstep[OF a_ndef[OF variable_right(5)] _ Rfun])
        fix x
        from sigma[unfolded normal_F_subst_def] have NF: "\<sigma> x \<in> NF_trs R" by auto
        thus "\<sigma> x \<in> NF (qrstep nfs Q R)" using NF_anti_mono[OF qrstep_mono[of R R Q "{}" nfs]] by auto
      qed
      with steps
      have id: "\<sigma> x = q \<cdot> \<sigma>" unfolding NF_def by (induct, auto)
      have subset: "vars_term q \<subseteq> set (vars_term_list q)" by auto
      note subst = subst[OF subset id, unfolded variable_right(3)[symmetric]]
      from valid[OF sigma, unfolded subst normalize_cc[of _ c] variable_right(1)] phi
      show "\<sigma> \<Turnstile> \<psi>" by auto
    qed
  next
    case (simplify_condition c \<phi> \<psi> ys cc \<phi>' \<psi>' \<sigma> vs c')
    hence c': "cc_valid c'" by auto
    show ?case
    proof (rule normalize_validI[OF simplify_condition(1)])
      fix \<delta>
      assume delta: "normal_F_subst \<delta>"
      assume "\<And>c'. c' \<in> set \<phi> \<Longrightarrow> \<delta> \<Turnstile> c'"
      with simplify_condition(2)
      have delta_bound: "\<delta> \<Turnstile> cc_bound ys cc" and delta_phi: "Ball (set \<phi>) (cc_models \<delta>)" by auto
      note delta_nf = delta[unfolded normal_F_subst_def]
      let ?\<delta>' = "\<sigma> \<circ>s \<delta>"
      {
        fix x
        assume "x \<notin> subst_domain \<sigma>"
        hence "\<sigma> x = Var x" unfolding subst_domain_def by auto
        hence "?\<delta>' x = \<delta> x" unfolding subst_compose_def by simp
      } note delta_id = this
      with simplify_condition(4) have delta_idd: "\<And> x. x \<notin> set ys \<Longrightarrow> \<delta> x = ?\<delta>' x" by force
      {
        fix t
        assume "t \<in> range ?\<delta>'"
        then obtain x where x: "?\<delta>' x = t" by auto
        have "funas_term t \<subseteq> F \<and> t \<in> NF_trs R"
        proof (cases "x \<in> subst_domain \<sigma>")
          case False
          from x delta_id[OF this] have "t \<in> range \<delta>" by auto
          with delta_nf show ?thesis by auto
        next
          case True
          with simplify_condition(4) have x_ys: "x \<in> set ys" by auto
          from x have t: "t = \<sigma> x \<cdot> \<delta>" unfolding subst_compose_def by auto
          from True have range: "\<sigma> x \<in> subst_range \<sigma>" by simp 
          from range simplify_condition(6) have F: "funas_term (\<sigma> x) \<subseteq> F" by force
          from range simplify_condition(6) have ndef: "\<And> fn. fn \<in> funas_term (\<sigma> x) \<Longrightarrow> \<not> defined R fn" by force
          show ?thesis unfolding t
          proof
            show "funas_term (\<sigma> x \<cdot> \<delta>) \<subseteq> F" using F delta_nf unfolding funas_term_subst by auto
          next
            have "\<sigma> x \<cdot> \<delta> \<in> NF (qrstep nfs {} R)"
            proof (rule NF_subst_qrstep[OF _ _ Rfun])
              fix fn
              assume "fn \<in> funas_term (\<sigma> x)"
              from ndef[OF this] show "\<not> defined (applicable_rules {} R) fn" unfolding applicable_rules_def applicable_rule_def by auto
            next
              fix y
              assume "y \<in> vars_term (\<sigma> x)"
              show "\<delta> y \<in> NF (qrstep nfs {} R)" using delta_nf by auto
            qed
            thus "\<sigma> x \<cdot> \<delta> \<in> NF_trs R" by auto
          qed
        qed
      }
      hence nF: "normal_F_subst ?\<delta>'" unfolding normal_F_subst_def by auto
      from cc_models_boundE[OF delta_idd delta_bound nF]
      have "?\<delta>' \<Turnstile> cc" .
      from this[unfolded normalize_cc[of _ cc] simplify_condition(3)]
      have "?\<delta>' \<Turnstile> CC_impl \<phi>' \<psi>'" .
      from this[unfolded cc_models_subst[OF simplify_condition(5), symmetric]]
      have delta_imp: "\<delta> \<Turnstile> CC_impl \<phi>' \<psi>' \<cdot>cc (\<sigma>,vs)" .
      {
        fix c
        assume "c \<in> set \<phi>'"
        from simplify_condition(7)[OF this] obtain d where d: "d \<in> set \<phi>"
          and imp: "d \<longrightarrow>cc c \<cdot>cc (\<sigma>,vs)" by auto
        from d delta_phi have "\<delta> \<Turnstile> d" by auto
        from cc_implies_substE[OF imp this]
        have "\<delta> \<Turnstile> c \<cdot>cc (\<sigma>,vs)" .
      }
      with delta_imp have "\<delta> \<Turnstile> \<psi>' \<cdot>cc (\<sigma>,vs)" by auto
      with delta_phi
      have "Ball (set (\<psi>' \<cdot>cc (\<sigma>,vs) # \<phi>)) (cc_models \<delta>)" by auto
      with cc_validE[OF cc_impliesE[OF simplify_condition(8) c'] delta]
      show "\<delta> \<Turnstile> \<psi>" by auto
    qed
  next
    case (fun_arg_into_var c \<phi> \<psi> c' p x f bef aft q)
    hence c': "cc_valid c'" by auto
    note valid = cc_validE[OF cc_impliesE[OF fun_arg_into_var(2) c']]
    note c = fun_arg_into_var(1)
    from fun_arg_into_var(1,3) have "x \<notin> vars_cc (CC_impl \<phi> \<psi>)" using vars_cc_normalize[of c] by auto
    note x_fresh = this[simplified]
    from fun_arg_into_var(4) x_fresh have "x \<notin> vars_cc (CC_rewr (Fun f (bef @ p # aft)) q)" by auto
    note x_fresh' = this[simplified]
    let ?t = "\<lambda> x. Fun f (bef @ x # aft)"
    show ?case
    proof (rule normalize_validI[OF c])
      fix \<sigma>
      assume sigma: "normal_F_subst \<sigma>"
      assume phi: "\<And>c'. c' \<in> set \<phi> \<Longrightarrow> \<sigma> \<Turnstile> c'"
      from phi[OF fun_arg_into_var(4)]
      have rewr: "\<sigma> \<Turnstile> CC_rewr (?t p) q" .
      let ?tsig = "\<lambda> x. Fun f (map (\<lambda> t. t \<cdot> \<sigma>) bef @ x # map (\<lambda> t. t \<cdot> \<sigma>) aft)"
      from rewr have SN: "m_SN (?tsig (p \<cdot> \<sigma>))" by simp
      from rewr have steps: "(?tsig (p \<cdot> \<sigma>), q \<cdot> \<sigma>) \<in> (qrstep nfs Q R)^*" by simp
      from rewr have NF: "q \<cdot> \<sigma> \<in> NFQ" by simp
      from normalize_subterm_qrsteps[OF _ steps NF, of "<length bef>"]
      obtain u where steps1: "(p \<cdot> \<sigma>, u) \<in> (qrstep nfs Q R)^*" and u: "u \<in> NFQ"
        and steps2: "(?tsig u, q \<cdot> \<sigma>) \<in> (qrstep nfs Q R)^*" by (auto simp: nth_append)        
      let ?\<delta> = "\<sigma>(x := u)"
      have p: "p \<cdot> ?\<delta> = p \<cdot> \<sigma>" by (rule term_subst_eq, insert x_fresh', auto)
      have q: "q \<cdot> ?\<delta> = q \<cdot> \<sigma>" by (rule term_subst_eq, insert x_fresh', auto)
      let ?m = "map (\<lambda>t. t \<cdot> \<sigma>)"
      let ?mu = "map (\<lambda>t. t \<cdot> \<sigma>(x := u))"
      have "(?t (Var x) \<cdot> \<sigma>(x := u) = ?tsig u) = (?mu (bef @ aft) = ?m (bef @ aft))" by simp
      moreover have "..."
        by (rule map_cong[OF refl term_subst_eq], insert x_fresh', auto)
      ultimately have tu: "?t (Var x) \<cdot> \<sigma>(x := u) = ?tsig u" by simp
      {
        assume m
        with SN[unfolded m_SN_def] have SN: "SN_on (qrstep nfs Q R) {?tsig (p \<cdot> \<sigma>)}" by blast
        from SN_imp_SN_arg_gen[OF ctxt_closed_qrstep SN, of "p \<cdot> \<sigma>"] have SNp: "SN_on (qrstep nfs Q R) {p \<cdot> \<sigma>}" by auto
        from steps_preserve_SN_on[OF ctxt_closedE[OF ctxt_closed_rtrancl[OF ctxt_closed_qrstep] steps1], of "More f (?m bef) Hole (?m aft)"]
        SN have SNu: "SN_on (qrstep nfs Q R) {?tsig u}" by auto
        note SNp SNu
      }
      hence SNp: "m_SN (p \<cdot> \<sigma>)" and SNu: "m_SN (?tsig u)" unfolding m_SN_def by auto
      {
        fix c
        assume "c \<in> set \<phi> \<union> {\<psi>}"
        with x_fresh have x: "x \<notin> vars_cc c" by auto
        have "\<sigma> \<Turnstile> c = ?\<delta> \<Turnstile> c"
          by (rule cc_models_vars, insert x, auto)
      } note id = this
      have rewr1: "?\<delta> \<Turnstile> CC_rewr p (Var x)" unfolding cc_models.simps p using u SNp steps1 by auto
      have rewr2: "?\<delta> \<Turnstile> CC_rewr (?t (Var x)) q" unfolding cc_models.simps tu q using steps2 NF SNu by auto
      have phi: "Ball (set \<phi>) (cc_models ?\<delta>)" using phi id by auto
      from NFQ u have NF: "u \<in> NF_trs R" by auto
      note sigma = sigma[unfolded normal_F_subst_def]
      have uF: "funas_term u \<subseteq> F"
        by (rule qrsteps_funas_term[OF wwf RF _ steps1], unfold funas_term_subst, insert sigma fun_arg_into_var(5), auto)
      {
        fix y
        have F: "funas_term (?\<delta> y) \<subseteq> F" using sigma uF by auto
        have NF: "?\<delta> y \<in> NF_trs R" using NF sigma by auto
        note F NF
      }
      hence delta: "normal_F_subst ?\<delta>" unfolding normal_F_subst_def by auto
      from id
      have "\<sigma> \<Turnstile> \<psi> = ?\<delta> \<Turnstile> \<psi>" by auto
      also have "..." using valid[OF delta] rewr1 rewr2 phi by auto
      finally show "\<sigma> \<Turnstile> \<psi>" .
    qed
  next
    case (induction cc \<phi>' \<psi> f xs q \<phi> cs)
    note cc = induction(1)
    note \<phi> = induction(4)
    note distinct = induction(6)
    let ?xs = "map Var xs :: ('f,'v)term list"
    let ?\<phi> = "CC_impl (CC_rewr (Fun f ?xs) q # \<phi>) \<psi>"
    let ?P = "\<lambda> \<delta>. normal_F_subst \<delta> \<longrightarrow> \<delta> \<Turnstile> ?\<phi>"
    let ?UR = "UR (Fun f ?xs)"
    from qrstep_mono[OF UR_sub subset_refl] have UR_R: "qrstep nfs Q ?UR \<subseteq> qrstep nfs Q R" .
    hence step_UR: "\<And> p. p \<in> qrstep nfs Q ?UR \<Longrightarrow> p \<in> qrstep nfs Q R" by auto
    from rtrancl_mono[OF UR_R] have steps_UR: "\<And> p. p \<in> (qrstep nfs Q ?UR)^* \<Longrightarrow> p \<in> (qrstep nfs Q R)^*" by auto
    {
      fix \<delta>
      assume delta: "normal_F_subst \<delta>"
      def dist \<equiv> "distance (qrstep nfs Q ?UR) (NF_terms Q)"
      def SNrel1 \<equiv> "inv_image (restrict_SN (qrstep nfs Q R O op \<unrhd>) (qrstep nfs Q R)) (\<lambda> \<delta>. (Fun f ?xs) \<cdot> \<delta>)"
      def SNrel2 \<equiv> "inv_image {(x,y). x > (y :: nat)} (\<lambda> \<delta>. dist (Fun f ?xs \<cdot> \<delta>))"
      def SNrel \<equiv> "if m then SNrel1 else SNrel2"
      have SN_rel: "SN SNrel"
      proof (cases m)
        case True
        have "SN SNrel1" unfolding SNrel1_def
        proof (rule SN_inv_image[OF SN_subset[OF SN_restrict_SN_idemp]])
          let ?R = "(op \<rhd> \<union> qrstep nfs Q R)^+"
          note d = restrict_SN_def
          show "restrict_SN (qrstep nfs Q R O op \<unrhd>) (qrstep nfs Q R) \<subseteq> restrict_SN ?R ?R" (is "?l \<subseteq> ?r")
          proof
            fix s t
            assume "(s,t) \<in> ?l"
            from this[unfolded d] have step: "(s,t) \<in> qrstep nfs Q R O op \<unrhd>" and SN: "SN_on (qrstep nfs Q R) {s}" by auto
            from step have "(s,t) \<in> qrstep nfs Q R \<union> qrstep nfs Q R O op \<rhd>" by auto
            hence step: "(s,t) \<in> ?R" by regexp
            with SN_on_trancl[OF SN_on_qrstep_imp_SN_on_supt_union_qrstep[OF SN]] show "(s,t) \<in> ?r" unfolding d by auto
          qed
        qed
        thus ?thesis unfolding SNrel_def using True by simp
      next
        case False
        have id: "{(x, y). y < x} = {(x, y). y > (x :: nat)}^-1" by auto
        have "SN SNrel2" unfolding SNrel2_def id
          by (rule SN_inv_image, rule wf_imp_SN, insert wf_less, simp)
        thus ?thesis unfolding SNrel_def using False by simp
      qed
      have "\<delta> \<Turnstile> ?\<phi>"
      proof (induct \<delta> rule: SN_induct[where P = ?P, rule_format, OF SN_rel])
        show "normal_F_subst \<delta>" by fact
      next
        fix \<delta>
        assume delta: "normal_F_subst \<delta>"
        assume IH: "\<And> \<tau>. (\<delta>, \<tau>) \<in> SNrel \<Longrightarrow> normal_F_subst \<tau> \<Longrightarrow> \<tau> \<Turnstile> ?\<phi>"
        show "\<delta> \<Turnstile> ?\<phi>" unfolding cc_models.simps(2)
        proof
          assume conj: "Ball (set (CC_rewr (Fun f ?xs) q # \<phi>)) (cc_models \<delta>)"
          hence rewr: "\<delta> \<Turnstile> CC_rewr (Fun f ?xs) q" and phi: "Ball (set \<phi>) (cc_models \<delta>)" by auto
          from rewr have steps: "(Fun f ?xs \<cdot> \<delta>, q \<cdot> \<delta>) \<in> (qrstep nfs Q R)^*" and nf_q: "q \<cdot> \<delta> \<in> NFQ" 
            and SN: "m_SN (Fun f ?xs \<cdot> \<delta>)" by auto
          {
            fix u R'
            assume u: "(Fun f ?xs \<cdot> \<delta>, u) \<in> qrstep nfs Q R'" and sub: "R' \<subseteq> R"
            have "(Fun f ?xs \<cdot> \<delta>, u) \<notin> nrqrstep nfs Q R'"
            proof
              assume "(Fun f ?xs \<cdot> \<delta>, u) \<in> nrqrstep nfs Q R'"
              from nrqrstep_imp_arg_qrstep[OF this] obtain i where 
                "(\<delta> (xs ! i), args u ! i) \<in> qrstep nfs Q R'" by auto
              with qrstep_mono[OF sub, of Q "{}" nfs] have "(\<delta> (xs ! i), args u ! i) \<in> rstep R" by auto
              moreover from delta[unfolded normal_F_subst_def] have "\<delta> (xs ! i) \<in> NF_trs R" by auto
              ultimately show False by auto
            qed            
            with u[unfolded qrstep_iff_rqrstep_or_nrqrstep] 
            have "(Fun f ?xs \<cdot> \<delta>, u) \<in> rqrstep nfs Q R'" by auto
          } note rqrstep = this
          have neq: "Fun f ?xs \<cdot> \<delta> \<noteq> q \<cdot> \<delta>"
          proof
            assume "Fun f ?xs \<cdot> \<delta> = q \<cdot> \<delta>"
            hence "\<delta> \<in> unifiers {(Fun f ?xs, q)}" unfolding unifiers_def by auto
            with mgu_complete[OF induction(2)] show False by auto
          qed
          {
            from rtranclD[OF steps, unfolded trancl_unfold_left] neq
            obtain u where step: "(Fun f ?xs \<cdot> \<delta>, u) \<in> qrstep nfs Q R" by auto
            from rqrstep[OF step subset_refl] have "(Fun f ?xs \<cdot> \<delta>, u) \<in> rqrstep nfs Q R" .
            from rqrstepE[OF this] have NFQ: "\<And> u. u \<lhd> Fun f ?xs \<cdot> \<delta> \<Longrightarrow> u \<in> NFQ" by metis
            {
              fix x
              assume "x \<in> vars_term (Fun f ?xs)"
              hence "Var x \<lhd> Fun f ?xs" by auto
              from NFQ[OF supt_stable[OF this]] have "\<delta> x \<in> NFQ" by auto
            }
            hence "\<delta> ` vars_term (Fun f ?xs) \<subseteq> NFQ" by auto
            from UR[OF this steps] have "(Fun f ?xs \<cdot> \<delta>, q \<cdot> \<delta>) \<in> (qrstep nfs Q ?UR)^*" .
          } note steps = this
          have "\<exists> u d. (Fun f ?xs \<cdot> \<delta>, u) \<in> qrstep nfs Q ?UR \<and> (u,q \<cdot> \<delta>) \<in> (qrstep nfs Q ?UR)^^d \<and> (m \<or> dist (Fun f ?xs \<cdot> \<delta>) = Suc d)"
          proof (cases m)
            case True
            from rtranclD[OF steps, unfolded trancl_unfold_left] neq
            obtain u where step: "(Fun f ?xs \<cdot> \<delta>, u) \<in> qrstep nfs Q ?UR" and steps: "(u,q \<cdot> \<delta>) \<in> (qrstep nfs Q ?UR)^*" by auto
            from step True rtrancl_imp_relpow[OF steps] show ?thesis by auto
          next
            case False
            from False induction(7) have CR: "CR (rstep ?UR)" by auto
            from CR_imp_UNF_inn[OF NFQ UR_sub CR] have "UNF_r_q (qrstep nfs Q ?UR) NFQ" .
            from distance_NF_UNF[OF subset_refl this nf_q steps]
            have "(Fun f ?xs \<cdot> \<delta>, q \<cdot> \<delta>) \<in> qrstep nfs Q ?UR ^^ distance (qrstep nfs Q ?UR) NFQ (Fun f ?xs \<cdot> \<delta>)" .
            from distance_Suc_relpow[OF nf_q this neq]
            show ?thesis unfolding dist_def by auto
          qed
          then obtain u d where step: "(Fun f ?xs \<cdot> \<delta>, u) \<in> qrstep nfs Q ?UR"
            and steps_d: "(u,q \<cdot> \<delta>) \<in> (qrstep nfs Q ?UR)^^d" and dist: "m \<or> dist (Fun f ?xs \<cdot> \<delta>) = Suc d" by auto
          from relpow_imp_rtrancl[OF steps_d] have steps: "(u,q \<cdot> \<delta>) \<in> (qrstep nfs Q ?UR)^*" .
          from SN step_preserves_SN_on[OF step_UR[OF step]] have SN_u: "m_SN u" unfolding m_SN_def by auto
          from rqrstep[OF step UR_sub]
          have "(Fun f ?xs \<cdot> \<delta>, u) \<in> rqrstep nfs Q ?UR" .
          from rqrstepE[OF this] obtain l r \<sigma>'' where 
            nf: "\<forall>u\<lhd>l \<cdot> \<sigma>''. u \<in> NFQ" and lr_U: "(l, r) \<in> ?UR" and fl: "Fun f ?xs \<cdot> \<delta> = l \<cdot> \<sigma>''" and u: "u = r \<cdot> \<sigma>''" and nfs: "NF_subst nfs (l, r) \<sigma>'' Q" .
          from rqrstepI[OF nf _ fl u nfs] have rstep: "(Fun f ?xs \<cdot> \<delta>, u) \<in> rqrstep nfs Q {(l,r)}" by auto
          from lr_U UR_sub have lr: "(l,r) \<in> R" by auto
          from only_applicable_rules[OF nf] lr wwf UR_sub have vars: "vars_term r \<subseteq> vars_term l" unfolding wwf_qtrs_def by auto
          from Rfun lr fl obtain ls where l: "l = Fun f ls" by (cases l, auto)
          from fl l have id: "map \<delta> xs = map (\<lambda> t. t \<cdot> \<sigma>'') ls" by (simp add: comp_def)
          from arg_cong[OF id, of length] have len: "length ls = length xs" by auto
          from induction(5)[OF len[symmetric] lr[unfolded l]] obtain c ls' r' rs_ys_list
            where eq: "(Fun f ls, r) =v (Fun f ls', r')" and vars_cc: "vars_rule (Fun f ls', r') \<inter> vars_cc ?\<phi> = {}"
            and c: "c \<longrightarrow>cc cc_rule_constraint f ls' r' q xs \<phi> \<psi> rs_ys_list" and c_cs: "c \<in> set cs" 
            and pre: "(\<forall>(r, ys)\<in>set rs_ys_list. root r = Some (f,length xs) \<and> r' \<unrhd> r \<and> vars_term r' \<inter> set ys = {} \<and> (\<forall>fn\<in>funas_args_term r. \<not> defined R fn))" by auto
          from rqrstep_rename_vars[of "{(Fun f ls,r)}" "{(Fun f ls',r')}" nfs Q] rstep[unfolded l] eq
          have "(Fun f ?xs \<cdot> \<delta>, u) \<in> rqrstep nfs Q {(Fun f ls',r')}" by auto
          from rqrstepE[OF this] obtain \<sigma>'' where 
            nf: "\<forall>u\<lhd>Fun f ls' \<cdot> \<sigma>''. u \<in> NFQ" and fl: "Fun f ?xs \<cdot> \<delta> = Fun f ls' \<cdot> \<sigma>''" and u: "u = r' \<cdot> \<sigma>''" and nfs: "NF_subst nfs (Fun f ls', r') \<sigma>'' Q" by auto
          from fl l have id: "map \<delta> xs = map (\<lambda> t. t \<cdot> \<sigma>'') ls'" by (simp add: comp_def)
          from arg_cong[OF id, of length] have len: "length ls' = length xs" by auto
          from eq_rule_mod_vars_var_cond[OF eq vars[unfolded l]] have vars: "vars_term r' \<subseteq> vars_term (Fun f ls')"  .
          def \<sigma> \<equiv> "mk_subst Var (zip xs ls')"
          have "ls' = map id ls'" by simp
          also have "... = map \<sigma> xs"
          proof (rule nth_map_conv[OF len], intro allI impI)
            fix i
            assume i: "i < length ls'"
            have "ls' ! i = \<sigma> (xs ! i)" unfolding \<sigma>_def using mk_subst_distinct[OF distinct len i[unfolded len]]
              by simp
            thus "id (ls' ! i) = \<sigma> (xs ! i)" by simp
          qed
          finally have ls: "ls' = map \<sigma> xs" .
          from id[unfolded ls] have "map \<delta> xs = map (\<lambda> x. \<sigma> x \<cdot> \<sigma>'') xs" by auto
          hence delta_sigma: "\<And> x. x \<in> set xs \<Longrightarrow> \<delta> x = \<sigma> x \<cdot> \<sigma>''" by auto
          def \<sigma>' \<equiv> "\<lambda> x. if x \<in> vars_term (Fun f ls') then \<sigma>'' x else \<delta> x"
          {
            fix t :: "('f,'v)term"
            assume sub: "vars_term t \<subseteq> vars_term (Fun f ls')"
            have "t \<cdot> \<sigma>'' = t \<cdot> \<sigma>'" unfolding \<sigma>'_def
              by (rule term_subst_eq, insert sub, auto)
          } note conv1 = this
          have delta_sigma': "\<And> x.  x \<notin> vars_term (Fun f ls') \<Longrightarrow> \<delta> x = (\<sigma> \<circ>s \<sigma>') x"
          proof -
            fix x
            assume x: "x \<notin> vars_term (Fun f ls')"
            have "\<sigma> x \<cdot> \<sigma>' = \<delta> x"
            proof (cases "x \<in> set xs")
              case True
              then obtain i where x: "x = xs ! i" and i: "i < length xs" unfolding set_conv_nth by auto
              have "\<sigma> x = ls' ! i" using arg_cong[OF ls, of "\<lambda> xs. xs ! i"] x i by auto
              with i len have mem: "\<sigma> x \<in> set ls'" by auto
              show ?thesis unfolding delta_sigma[OF True]
                by (rule conv1[symmetric], insert mem, auto)
            next
              case False
              from mk_subst_not_mem[OF this]
              have "\<sigma> x = Var x" unfolding \<sigma>_def . 
              hence "\<sigma> x \<cdot> \<sigma>' = \<sigma>' x" by simp
              also have "... = \<delta> x" unfolding \<sigma>'_def using x by simp
              finally show ?thesis by simp
            qed
            thus "\<delta> x = (\<sigma> \<circ>s \<sigma>') x" unfolding subst_compose_def by auto
          qed
          def \<delta>' \<equiv> "\<lambda> x. if x \<in> vars_term (Fun f ls') then (\<sigma> \<circ>s \<sigma>') x else \<delta> x"
          have delta': "\<delta>' = \<sigma> \<circ>s \<sigma>'" unfolding \<delta>'_def using delta_sigma' by auto
          {
            fix c :: "('f,'v)cond_constraint"
            assume "vars_cc c \<subseteq> vars_cc ?\<phi>"
            with vars_cc have disj: "vars_cc c \<inter> vars_term (Fun f ls') = {}" unfolding vars_rule_def by auto
            have "\<delta> \<Turnstile> c = \<delta>' \<Turnstile> c"
              by (rule cc_models_vars, insert disj \<delta>'_def, auto)
          } note delta_switch = this
          from delta_switch phi have phi: "Ball (set \<phi>) (cc_models \<delta>')" unfolding \<phi> by auto
          {
            fix x
            have "funas_term (\<sigma>' x) \<subseteq> F \<and> \<sigma>' x \<in> NF_trs R"
            proof (cases "x \<in> vars_term (Fun f ls')")
              case True
              hence sigma': "\<sigma>' x = \<sigma>'' x" unfolding \<sigma>'_def by simp
              from True obtain l where l: "l \<in> set ls'" and x: "x \<in> vars_term l" by auto
              from l[unfolded set_conv_nth] obtain i where i: "i < length ls'" and l: "l = ls' ! i" by auto
              from arg_cong[OF id, of "\<lambda> ls. ls ! i"] i len l have "l \<cdot> \<sigma>'' = \<delta> (xs ! i)" by auto
              with delta[unfolded normal_F_subst_def] have F: "funas_term (l \<cdot> \<sigma>'') \<subseteq> F" and NF: "l \<cdot> \<sigma>'' \<in> NF_trs R" by auto
              with l x have F: "funas_term (\<sigma>' x) \<subseteq> F" unfolding sigma' funas_term_subst by auto
              from x have "l \<unrhd> Var x" by auto
              hence "l \<cdot> \<sigma>'' \<unrhd> Var x \<cdot> \<sigma>''" by auto
              from NF_subterm[OF NF this] F
              show ?thesis unfolding sigma' by auto
            next
              case False
              hence "\<sigma>' x = \<delta> x" unfolding \<sigma>'_def by simp
              thus ?thesis using delta unfolding normal_F_subst_def by auto
            qed
          }
          hence sigmaF: "normal_F_subst \<sigma>'" unfolding normal_F_subst_def by auto
          have q: "q \<cdot> \<delta> = q \<cdot> \<delta>'"
            by (rule term_subst_eq, unfold \<delta>'_def, insert vars_cc, auto simp: vars_rule_def cc \<phi>)
          have u: "u = r' \<cdot> \<sigma>'" unfolding u
          proof (rule term_subst_eq)
            fix x
            assume "x \<in> vars_term r'"
            with vars have "x \<in> vars_term (Fun f ls')" by auto
            thus "\<sigma>'' x = \<sigma>' x" unfolding \<sigma>'_def by auto
          qed
          have sigma'_rewr: "\<sigma>' \<Turnstile> CC_rewr r' (q \<cdot> \<sigma>)" 
            unfolding cc_models.simps using steps_UR[OF steps] nf_q SN_u unfolding q delta' u by auto
          from cc_validE[OF _ sigmaF] c_cs induction(8) have sigma'_c: "\<sigma>' \<Turnstile> c" by auto        
          def vs \<equiv> "vars_subst_range_impl (zip xs ls')"
          have "vars_subst_range \<sigma> \<subseteq> set vs" unfolding vs_def \<sigma>_def by simp
          note cc_sigma = cc_models_subst[OF this]
          note c = c[unfolded cc_rule_constraint_def Let_def \<sigma>_def[symmetric] vs_def[symmetric]]
          have "Ball (set (cc_ih_prems f q xs \<phi> \<psi> rs_ys_list)) (cc_models \<sigma>')" unfolding cc_models.simps
          proof
            fix c
            assume "c \<in> set (cc_ih_prems f q xs \<phi> \<psi> rs_ys_list)"
            note c = this[unfolded cc_ih_prems_def Let_def]
            from c obtain rr ys where c: "c =  cc_bound ys
                            (CC_impl (CC_rewr rr (q \<cdot> mk_subst Var (zip xs (args rr))) #
                                               map (\<lambda>c. c \<cdot>cc (mk_subst Var (zip xs (args rr)), vars_subst_range_impl (zip xs (args rr)))) \<phi>)
                              (\<psi> \<cdot>cc (mk_subst Var (zip xs (args rr)), vars_subst_range_impl (zip xs (args rr)))))" and mem: "(rr,ys) \<in> set rs_ys_list" by force
            from mem pre obtain rs where rr: "rr = Fun f rs" "args rr = rs" by (cases rr, auto)
            note c = c[unfolded rr]
            from mem pre rr have
              lenrs: "length rs = length xs" and subt: "r' \<unrhd> Fun f rs" and rvars: "vars_term r' \<inter> set ys = {}" 
                and ndef: "\<forall> fn \<in> funas_args_term (Fun f rs). \<not> defined R fn" by auto
            from supteq_imp_vars_subset[OF subt] rvars have rs_vars: "vars_term (Fun f rs) \<inter> set ys = {}" by auto
            def vs' \<equiv> "vars_subst_range_impl (zip xs rs)"
            def \<mu> \<equiv> "mk_subst Var (zip xs rs)"
            from c vs'_def \<mu>_def
            have c: "c = cc_bound ys (CC_impl (CC_rewr (Fun f rs) (q \<cdot> \<mu>) # map (\<lambda>c. c \<cdot>cc (\<mu>, vs')) \<phi>) (\<psi> \<cdot>cc (\<mu>, vs')))" by simp
            have "rs = map id rs" by simp
            also have "... = map \<mu> xs"
            proof (rule nth_map_conv[OF lenrs], intro allI impI)
              fix i
              assume i: "i < length rs"
              have "rs ! i = \<mu> (xs ! i)" unfolding \<mu>_def using mk_subst_distinct[OF distinct lenrs i[unfolded lenrs]]
                by simp
              thus "id (rs ! i) = \<mu> (xs ! i)" by simp
            qed
            finally have "rs = map \<mu> xs" .
            hence rsmu: "Fun f rs = Fun f ?xs \<cdot> \<mu>" by simp
            have rs_mu: "cc_bound ys (CC_impl (CC_rewr (Fun f rs) (q \<cdot> \<mu>) # map (\<lambda>c. c \<cdot>cc (\<mu>, vs')) \<phi>) (\<psi> \<cdot>cc (\<mu>, vs')))
              = cc_bound ys ((CC_impl (CC_rewr (Fun f ?xs) q # \<phi>) \<psi>) \<cdot>cc (\<mu>,vs'))" unfolding rsmu by simp
            have "\<sigma>' \<Turnstile> cc_bound ys (CC_impl (CC_rewr (Fun f rs) (q \<cdot> \<mu>) # map (\<lambda>c. c \<cdot>cc (\<mu>, vs')) \<phi>) (\<psi> \<cdot>cc (\<mu>, vs')))" unfolding rs_mu
            proof (rule cc_models_boundI)
              fix \<tau>
              assume tau_sigma: "\<And>x. x \<notin> set ys \<Longrightarrow> \<tau> x = \<sigma>' x"
                and tau: "\<And>x. x \<in> set ys \<Longrightarrow> funas_term (\<tau> x) \<subseteq> F \<and> \<tau> x \<in> NF_trs R"
              {
                fix x
                have "funas_term (\<tau> x) \<subseteq> F \<and> \<tau> x \<in> NF_trs R"
                using tau_sigma[of x] tau[of x] sigmaF[unfolded normal_F_subst_def]
                  by (cases "x \<in> set ys", force+)
              } note tauF = this
              hence tau_F: "normal_F_subst \<tau>" unfolding normal_F_subst_def by auto
              have vs': "vars_subst_range \<mu> \<subseteq> set vs'" unfolding vs'_def \<mu>_def by auto
              show "\<tau> \<Turnstile> (CC_impl (CC_rewr (Fun f ?xs) q # \<phi>) \<psi>) \<cdot>cc (\<mu>, vs')" 
                unfolding cc_models_subst[OF vs']
              proof (rule IH)
                {
                  fix x
                  let ?x = "\<mu> x \<cdot> \<tau>"
                  have "funas_term ?x \<subseteq> F \<and> ?x \<in> NF_trs R"
                  proof (cases "x \<in> set xs")
                    case False
                    from mk_subst_not_mem[OF this]
                    have "\<mu> x = Var x" unfolding \<mu>_def .
                    with tauF[of x] show ?thesis by simp
                  next
                    case True
                    then obtain i where i: "i < length xs" and x: "x = xs ! i" unfolding set_conv_nth by auto
                    from mk_subst_distinct[OF distinct lenrs i[unfolded lenrs]]
                    have "\<mu> x = rs ! i" unfolding \<mu>_def x .
                    with i lenrs have mu: "\<mu> x \<in> set rs" by auto
                    from eq_rule_mod_varsE[OF eq] obtain \<gamma> where r: "r = r' \<cdot> \<gamma>" by auto
                    have "funas_term (\<mu> x) \<subseteq> funas_term (Fun f rs)" using mu by auto
                    also have "\<dots> \<subseteq> funas_term r'" using subt by fastforce
                    also have "\<dots> \<subseteq> funas_term r" unfolding r funas_term_subst by auto
                    also have "\<dots> \<subseteq> F" using lr RF unfolding funas_trs_def funas_rule_def by force
                    finally have F: "funas_term ?x \<subseteq> F" using tauF unfolding funas_term_subst by auto
                    have "(?x \<in> NF_trs R) = (?x \<in> NF (qrstep nfs {} R))" by simp
                    also have "\<dots>"
                      by (rule NF_subst_qrstep[OF _ _ Rfun], insert tauF ndef mu, 
                        auto simp: applicable_rules_def applicable_rule_def funas_args_term_def)
                    finally show ?thesis using F by auto
                  qed
                }
                thus "normal_F_subst (\<mu> \<circ>s \<tau>)" unfolding subst_compose_def normal_F_subst_def by auto
                have "(Fun f ?xs \<cdot> \<delta>, u) \<in> qrstep nfs Q R" using step_UR[OF step] .
                also have "u = r' \<cdot> \<sigma>'" unfolding u ..
                finally have step: "(Fun f ?xs \<cdot> \<delta>, Fun f rs \<cdot> \<sigma>') \<in> qrstep nfs Q R O op \<unrhd>" using supteq_stable[OF subt] by auto
                have "Fun f rs \<cdot> \<sigma>' = Fun f rs \<cdot> \<tau>"
                  by (rule term_subst_eq[OF tau_sigma[symmetric]], insert rs_vars, auto)
                also have "Fun f rs \<cdot> \<tau> = Fun f ?xs \<cdot> \<mu> \<cdot> \<tau>" unfolding rsmu by simp
                finally have id: "Fun f rs \<cdot> \<sigma>' = Fun f ?xs \<cdot> \<mu> \<circ>s \<tau>" by simp
                from step id have stepi: "(Fun f ?xs \<cdot> \<delta>, Fun f ?xs \<cdot> \<mu> \<circ>s \<tau>) \<in> qrstep nfs Q R O op \<unrhd>" by simp
                show "(\<delta>, \<mu> \<circ>s \<tau>) \<in> SNrel"
                proof (cases m)
                  case True
                  with SN have "SN_on (qrstep nfs Q R) {Fun f ?xs \<cdot> \<delta>}" unfolding m_SN_def by auto
                  hence "(\<delta>, \<mu> \<circ>s \<tau>) \<in> SNrel1" using stepi unfolding SNrel1_def inv_image_def restrict_SN_def by force
                  thus ?thesis using True unfolding SNrel_def by auto
                next
                  case False
                  with dist have dist: "dist (Fun f ?xs \<cdot> \<delta>) = Suc d" and SNrel: "SNrel = SNrel2" unfolding SNrel_def by auto
                  from steps_d[unfolded u] have steps_d: "(r' \<cdot> \<sigma>', q \<cdot> \<delta>) \<in> qrstep nfs Q ?UR ^^ d" .
                  have "dist (Fun f ?xs \<cdot> \<mu> \<circ>s \<tau>) = dist (Fun f rs \<cdot> \<sigma>')" unfolding id by simp
                  also have "\<dots> \<le> dist (r' \<cdot> \<sigma>')" unfolding dist_def
                    by (rule distance_subterm_innermost[OF nf_q steps[unfolded u] supteq_stable[OF subt]]) 
                  also have "\<dots> \<le> d" unfolding dist_def
                    by (rule distance_least[OF nf_q steps_d])
                  also have "\<dots> < dist (Fun f ?xs \<cdot> \<delta>)" unfolding dist by simp
                  finally have dist: "dist (Fun f ?xs \<cdot> \<mu> \<circ>s \<tau>) < dist (Fun f ?xs \<cdot> \<delta>)" .
                  show ?thesis unfolding SNrel SNrel2_def inv_image_def 
                    by (rule, unfold split, rule, unfold split, rule dist)
                qed
              qed
            qed
            thus "\<sigma>' \<Turnstile> c" unfolding c .
          qed
          with sigma'_rewr cc_implies_substE[OF c sigma'_c]
          have sigma'_impl: "\<sigma>' \<Turnstile> (CC_impl \<phi> \<psi>) \<cdot>cc (\<sigma>,vs)" by auto
          from sigma'_impl[unfolded cc_sigma] 
          have "\<delta>' \<Turnstile> CC_impl \<phi> \<psi>" unfolding delta' by auto
          with phi have "\<delta>' \<Turnstile> \<psi>" by auto
          with delta_switch[of \<psi>] show "\<delta> \<Turnstile> \<psi>" by simp
        qed
      qed
      hence "\<delta> \<Turnstile> CC_impl \<phi>' \<psi>" using induction(3-4) by auto
    }
    thus ?case unfolding cc_valid_def normalize_cc[of _ cc] cc by auto
  qed
qed
end

end
